from typing import List, Tuple, Any

import numpy as np

from nebullvm.config import NO_COMPILER_INSTALLATION

try:
    from openvino.tools.pot import DataLoader
    from openvino.tools.pot import IEEngine
    from openvino.tools.pot import load_model, save_model
    from openvino.tools.pot import compress_model_weights
    from openvino.tools.pot import create_pipeline
except ImportError:
    import cpuinfo

    if (
        "intel" in cpuinfo.get_cpu_info()["brand_raw"].lower()
        and not NO_COMPILER_INSTALLATION
    ):
        from nebullvm.installers.installers import install_openvino

        install_openvino()
        from openvino.tools.pot import DataLoader
        from openvino.tools.pot import IEEngine
        from openvino.tools.pot import load_model, save_model
        from openvino.tools.pot import compress_model_weights
        from openvino.tools.pot import create_pipeline
    else:
        DataLoader = object


class _CalibrationDataLoader(DataLoader):
    def __init__(
        self, input_data: List[Tuple[Any, ...]], input_names: List[str]
    ):
        self._input_data = input_data
        self._input_names = input_names

    def __len__(self):
        return len(self._input_data[0])

    def __getitem__(self, item):
        return (
            dict(zip(self._input_names, self._input_data[0][item])),
            self._input_data[1][item],
        )


def quantize_openvino(
    model_topology: str,
    model_weights: str,
    input_data: List[Tuple[np.ndarray, ...]],
    input_names: List[str],
) -> Tuple[str, str]:
    model_config = {
        "model_name": "model",
        "model": model_topology,
        "weights": model_weights,
    }

    # Engine config
    engine_config = {"device": "CPU"}

    algorithms = [
        {
            "name": "DefaultQuantization",
            "params": {
                "target_device": "ANY",
                "preset": "performance",
                "stat_subset_size": len(input_data),
            },
        }
    ]
    data_loader = _CalibrationDataLoader(
        input_data=input_data, input_names=input_names
    )
    model = load_model(model_config=model_config)
    engine = IEEngine(config=engine_config, data_loader=data_loader)
    pipeline = create_pipeline(algorithms, engine)
    compressed_model = pipeline.run(model=model)
    compress_model_weights(compressed_model)
    compressed_model_paths = save_model(
        model=compressed_model,
        save_path="quantized_model",
        model_name="quantized_model",
    )
    return (
        compressed_model_paths[0]["model"],
        compressed_model_paths[0]["weights"],
    )



from typing import Union, List, Tuple

import tensorflow as tf

from nebullvm.base import InputInfo, DataType


def get_outputs_sizes_tf(
    tf_model: Union[tf.Module, tf.keras.Model], input_tensors: List[tf.Tensor]
) -> List[Tuple[int, ...]]:
    outputs = tf_model(*input_tensors)
    if isinstance(outputs, tf.Tensor):
        return [tuple(outputs.shape)]
    return [tuple(x.shape) for x in outputs]


def create_model_inputs_tf(
    batch_size: int, input_infos: List[InputInfo]
) -> List[tf.Tensor]:
    return [
        tf.random_normal_initializer()(shape=(batch_size, *input_info.size))
        if input_info.dtype is DataType.FLOAT
        else tf.random.uniform(
            shape=(batch_size, *input_info.size),
            minval=input_info.min_value or 0,
            maxval=input_info.max_value or 100,
            dtype=tf.int32,
        )
        for input_info in input_infos
    ]


def run_tf_model(
    model: tf.Module, input_tensors: Tuple[tf.Tensor]
) -> Tuple[tf.Tensor]:
    pred = model.predict(*input_tensors)
    if isinstance(pred, tf.Module):
        pred = (pred,)
    return pred



from dataclasses import dataclass
from enum import Enum
from typing import Tuple, List, Dict, Union


class DataType(str, Enum):
    FLOAT = "float"
    INT = "int"


@dataclass
class DynamicAxisInfo:
    inputs: List[Dict[int, str]]
    outputs: List[Dict[int, str]]

    def dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith("_")
        }

    def retrieve_output_dim(
        self,
        input_shapes: List[Tuple[int, ...]],
        output_idx: int,
        dimension_idx: int,
        default_output_value: int,
    ) -> int:
        output_tag = self.outputs[output_idx][dimension_idx]
        for input_dict, input_shape in zip(self.inputs, input_shapes):
            for key, value in input_dict.items():
                if value == output_tag:
                    return input_shape[key]
        return default_output_value


class InputInfo:
    """Class for storing all the information needed for creating an input
    tensor for AI models.

    Attributes:
        size (tuple): Tuple with the input size (batch size excluded)
        dtype (str): Data type of the tensor.
        min_value (int or float, optional): Min value the tensor elements can
            have.
        max_value (int or float, optional): Max value the tensor elements can
            have.
    """

    def __init__(self, size: Tuple[int, ...], dtype: str, **extra_info):
        self.dtype = DataType(dtype)
        self.size = size
        self.__dict__.update(extra_info)

    def __getattr__(self, item):
        return self.__dict__.get(item)

    def dict(self):
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith("_")
        }


@dataclass
class ModelParams:
    batch_size: int
    input_infos: List[InputInfo]
    output_sizes: List[Tuple[int, ...]]
    dynamic_info: Union[DynamicAxisInfo, Dict] = None

    def __post_init__(self):
        if isinstance(self.dynamic_info, dict):
            self.dynamic_info = DynamicAxisInfo(**self.dynamic_info)
        self.input_infos = [
            InputInfo(**x) if isinstance(x, dict) else x
            for x in self.input_infos
        ]

    def dict(self):
        def recursively_dictionarize(element):
            if isinstance(element, list):
                element = [recursively_dictionarize(el) for el in element]
            elif hasattr(element, "dict"):
                element = element.dict()
            return element

        return {
            k: recursively_dictionarize(v)
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }

    @property
    def input_sizes(self):
        for input_info in self.input_infos:
            yield input_info.size


class DeepLearningFramework(Enum):
    PYTORCH = "torch"
    TENSORFLOW = "tensorflow"
    NUMPY = "numpy"


class ModelCompiler(Enum):
    TENSOR_RT = "tensor RT"
    OPENVINO = "openvino"
    APACHE_TVM = "tvm"
    ONNX_RUNTIME = "onnxruntime"
    TORCHVISION = "torchvision"
    TFLITE = "tflite"


class QuantizationType(Enum):
    DYNAMIC = "DYNAMIC"
    STATIC = "STATIC"
    QAT = "QAT"
    HALF = "HALF"



import warnings
from typing import Tuple, Callable, Any, List

import numpy as np

from nebullvm.base import QuantizationType
from nebullvm.inference_learners.base import BaseInferenceLearner
from nebullvm.measure import compute_relative_difference


def check_precision(
    optimized_learner: BaseInferenceLearner,
    input_data: List[Tuple[Any, ...]],
    base_outputs_list: List[Tuple[Any, ...]],
    perf_loss_ths: float,
    metric_func: Callable = None,
    ys: List = None,
    aggregation_func: Callable = np.mean,
) -> bool:
    metric_func = metric_func or compute_relative_difference
    relative_differences = []
    if ys is None:
        ys = [None] * len(input_data)
    for inputs, base_outputs, y in zip(input_data, base_outputs_list, ys):
        opt_outputs = optimized_learner(*inputs)
        relative_difference = max(
            metric_func(base_output, opt_output, y)
            for base_output, opt_output in zip(base_outputs, opt_outputs)
        )
        relative_differences.append(relative_difference)
    relative_difference = aggregation_func(relative_differences)
    return relative_difference <= perf_loss_ths


def check_quantization(
    quantization_type: QuantizationType, perf_loss_ths: float
):
    if quantization_type is None and perf_loss_ths is not None:
        raise ValueError(
            "When a quantization threshold is given it is necessary to "
            "specify the quantization algorithm too."
        )
    if quantization_type is not None and perf_loss_ths is None:
        warnings.warn(
            "Got a valid quantization type without any given quantization "
            "threshold. The quantization step will be ignored."
        )



from typing import List, Tuple, Any

import numpy as np
import onnx
import tensorflow as tf
import torch

from nebullvm.base import InputInfo, DataType, DeepLearningFramework


def convert_to_numpy(tensor: Any):
    if isinstance(tensor, torch.Tensor):
        tensor = tensor.cpu().detach().numpy()
    elif isinstance(tensor, tf.Tensor):
        tensor = tensor.numpy()
    else:
        if not isinstance(tensor, np.ndarray):
            raise TypeError(f"Unsupported data type: {type(tensor)}")
    return tensor


def convert_to_target_framework(
    tensor: np.ndarray, framework: DeepLearningFramework
) -> Any:
    if framework is DeepLearningFramework.PYTORCH:
        return torch.from_numpy(tensor)
    elif framework is DeepLearningFramework.TENSORFLOW:
        return tf.convert_to_tensor(tensor)
    else:
        return tensor


def get_input_names(onnx_model: str):
    model = onnx.load(onnx_model)
    input_all = [node.name for node in model.graph.input]
    return input_all


def get_output_names(onnx_model: str):
    model = onnx.load(onnx_model)
    output_all = [node.name for node in model.graph.output]
    return output_all


def run_onnx_model(
    onnx_model: str, input_tensors: List[np.ndarray]
) -> List[np.ndarray]:
    import onnxruntime as ort

    model = ort.InferenceSession(
        onnx_model, providers=["CPUExecutionProvider"]
    )
    inputs = {
        name: array
        for name, array in zip(get_input_names(onnx_model), input_tensors)
    }
    res = model.run(
        output_names=get_output_names(onnx_model), input_feed=inputs
    )
    return list(res)


def get_output_sizes_onnx(
    onnx_model: str, input_tensors: List[np.ndarray]
) -> List[Tuple[int, ...]]:
    res = run_onnx_model(onnx_model, input_tensors)
    sizes = [tuple(output.shape[1:]) for output in res]
    return sizes


def create_model_inputs_onnx(
    batch_size: int, input_infos: List[InputInfo]
) -> List[np.ndarray]:
    input_tensors = (
        np.random.randn(batch_size, *input_info.size).astype(np.float32)
        if input_info.dtype is DataType.FLOAT
        else np.random.randint(
            size=(batch_size, *input_info.size),
            low=input_info.min_value or 0,
            high=input_info.max_value or 100,
        )
        for input_info in input_infos
    )
    return list(input_tensors)



from pathlib import Path
from typing import Union, Iterable, Tuple, List

import cpuinfo
import numpy as np
import onnx
import torch
from torch.utils.data import DataLoader

from nebullvm.base import QuantizationType
from nebullvm.config import NO_COMPILER_INSTALLATION
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.transformations.precision_tfms import HalfPrecisionTransformation
from nebullvm.utils.onnx import get_input_names

try:
    from onnxmltools.utils.float16_converter import (
        convert_float_to_float16_model_path,
    )
    from onnxruntime.quantization import (
        QuantType,
        quantize_static,
        quantize_dynamic,
        CalibrationDataReader,
    )
except ImportError:
    if NO_COMPILER_INSTALLATION:
        QuantType = quantize_static = quantize_dynamic = None
        CalibrationDataReader = object
    else:
        from nebullvm.installers.installers import install_onnxruntime

        install_onnxruntime()
        from onnxruntime.quantization import (
            QuantType,
            quantize_static,
            quantize_dynamic,
            CalibrationDataReader,
        )


class _IterableCalibrationDataReader(CalibrationDataReader):
    def __init__(
        self,
        iterable_dataset: Union[Iterable[Tuple], List[Tuple]],
        input_names: List[str],
    ):
        self.iterable_dataset = iter(
            [
                {
                    input_name: value
                    for inputs in iterable_dataset
                    for input_name, value in zip(input_names, inputs)
                }
            ]
        )

    def get_next(self) -> dict:
        return next(self.iterable_dataset, None)

    @classmethod
    def from_dataloader(
        cls, dl: DataLoader, input_names: List[str], contains_y: bool = True
    ):
        iterable_ds = iter(
            inputs[:-1] if contains_y else inputs for inputs in dl
        )
        return cls(iterable_ds, input_names)


def _quantize_dynamic(model_path: str):
    model_path = Path(model_path)
    model_quant = model_path.parent / (model_path.stem + ".quant.onnx")
    quantize_dynamic(
        model_path,
        model_quant,
        weight_type=QuantType.QUInt8,
        optimize_model=False,
    )
    return str(model_quant)


def _get_quantization_type_for_static() -> Tuple[QuantType, QuantType]:
    """Returns the quantization types for activations and weights,
    depending on the underlying hardware
    """
    arch = cpuinfo.get_cpu_info()["arch"].lower()
    if torch.cuda.is_available():
        activation_type = weight_type = QuantType.QInt8
    elif "x86" in arch:
        cpu_raw_data = cpuinfo.get_cpu_info()["brand_raw"].lower()
        if "intel" in cpu_raw_data and "xeon" in cpu_raw_data:
            activation_type = QuantType.QUInt8
            weight_type = QuantType.QInt8
        else:
            activation_type = weight_type = QuantType.QUInt8
    else:
        activation_type = QuantType.QUInt8
        weight_type = QuantType.QInt8
    return activation_type, weight_type


def _quantize_static(
    model_path: str, input_data: List[Tuple[np.ndarray, ...]]
):
    model_path = Path(model_path)
    model_quant = model_path.parent / (model_path.stem + ".quant.onnx")
    inputs = input_data
    input_names = get_input_names(str(model_path))
    cdr = _IterableCalibrationDataReader(
        input_names=input_names, iterable_dataset=inputs
    )
    activation_type, weight_type = _get_quantization_type_for_static()
    quantize_static(
        model_path,
        Path(model_quant),
        cdr,
        activation_type=activation_type,
        weight_type=weight_type,
        optimize_model=False,
    )
    return str(model_quant)


def _convert_to_half_precision(
    model_path: str, input_tfms: MultiStageTransformation
):
    model_path = Path(model_path)
    model_quant = model_path.parent / (model_path.stem + "_fp16.onnx")
    new_onnx_model = convert_float_to_float16_model_path(model_path)
    input_tfms.append(HalfPrecisionTransformation())
    onnx.save(new_onnx_model, str(model_quant))
    return str(model_quant), input_tfms


def quantize_onnx(
    model_path: str,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data: List[Tuple[np.ndarray, ...]],
):
    if quantization_type is QuantizationType.STATIC:
        model_path = _quantize_static(model_path, input_data)
    elif quantization_type is QuantizationType.DYNAMIC:
        model_path = _quantize_dynamic(model_path)
    elif quantization_type is QuantizationType.HALF:
        model_path, input_tfms = _convert_to_half_precision(
            model_path, input_tfms
        )

    return model_path, input_tfms



from pathlib import Path
from typing import Tuple, Union, Optional

import torch

from nebullvm.base import ModelParams
from nebullvm.inference_learners import (
    PytorchBaseInferenceLearner,
    LearnerMetadata,
)
from nebullvm.transformations.base import MultiStageTransformation


class PytorchBackendInferenceLearner(PytorchBaseInferenceLearner):
    MODEL_NAME = "model_scripted.pt"

    def __init__(self, torch_model: torch.jit.ScriptModule, **kwargs):
        super().__init__(**kwargs)
        self.model = torch_model.eval()
        if torch.cuda.is_available():
            self.model.cuda()

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        device = input_tensors[0].device
        if torch.cuda.is_available():
            input_tensors = (t.cuda() for t in input_tensors)
        with torch.no_grad():
            res = self.model(*input_tensors)
            if not isinstance(res, tuple):
                res = res.to(device)
                return (res,)
            return tuple(out.to(device) for out in res)

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        self.model.save(path / self.MODEL_NAME)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        model = torch.jit.load(path / cls.MODEL_NAME)
        metadata = LearnerMetadata.read(path)
        return cls(
            torch_model=model,
            network_parameters=ModelParams(**metadata.network_parameters),
            input_tfms=metadata.input_tfms,
        )

    @classmethod
    def from_torch_model(
        cls,
        model: torch.nn.Module,
        network_parameters: ModelParams,
        input_tfms: Optional[MultiStageTransformation] = None,
    ):
        model_scripted = torch.jit.script(model)
        return cls(
            torch_model=model_scripted,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
        )



import shutil
from pathlib import Path
from typing import Tuple, Union, Dict, Type

import tensorflow as tf

from nebullvm.base import ModelParams
from nebullvm.config import TENSORFLOW_BACKEND_FILENAMES
from nebullvm.inference_learners import (
    TensorflowBaseInferenceLearner,
    LearnerMetadata,
)


class TensorflowBackendInferenceLearner(TensorflowBaseInferenceLearner):
    def __init__(self, tf_model: tf.Module, **kwargs):
        super(TensorflowBackendInferenceLearner, self).__init__(**kwargs)
        self.model = tf_model

    @tf.function(jit_compile=True)
    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        res = self.model.predict(*input_tensors)
        if not isinstance(res, tuple):
            return (res,)
        return res

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        self.model.save(path / TENSORFLOW_BACKEND_FILENAMES["tf_model"])

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.read(path)
        network_parameters = ModelParams(**metadata.network_parameters)
        input_tfms = metadata.input_tfms
        model = tf.keras.models.load_model(
            path / TENSORFLOW_BACKEND_FILENAMES["tf_model"]
        )
        return cls(
            tf_model=model,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
        )


class TFLiteBackendInferenceLearner(TensorflowBaseInferenceLearner):
    def __init__(self, tflite_file: str, **kwargs):
        super(TFLiteBackendInferenceLearner, self).__init__(**kwargs)
        self._tflite_file = self._store_file(tflite_file)
        self.interpreter = tf.lite.Interpreter(tflite_file)

    def run(self, *input_tensors: tf.Tensor):
        input_details = self.interpreter.get_input_details()
        output_details = self.interpreter.get_output_details()
        if self.network_parameters.dynamic_info:
            for i, (input_tensor, detail) in enumerate(
                zip(input_tensors, input_details)
            ):
                if input_tensor.shape != detail["shape"]:
                    self.interpreter.resize_tensor_input(i, input_tensor.shape)
        self.interpreter.allocate_tensors()
        for i, input_tensor in enumerate(input_tensors):
            self.interpreter.set_tensor(i, input_tensor)
        self.interpreter.invoke()
        return tuple(
            self.interpreter.get_tensor(output_detail["index"])
            for output_detail in output_details
        )

    def save(self, path: Union[str, Path], **kwargs):
        path = Path(path)
        metadata = LearnerMetadata.from_model(self, **kwargs)
        metadata.save(path)
        shutil.copy2(
            self._tflite_file,
            path / TENSORFLOW_BACKEND_FILENAMES["tflite_model"],
        )

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        path = Path(path)
        tflite_file = str(path / TENSORFLOW_BACKEND_FILENAMES["tflite_model"])
        metadata = LearnerMetadata.read(path)
        network_parameters = ModelParams(**metadata.network_parameters)
        input_tfms = metadata.input_tfms
        return cls(
            tflite_file=tflite_file,
            network_parameters=network_parameters,
            input_tfms=input_tfms,
        )


TF_BACKEND_LEARNERS_DICT: Dict[
    str,
    Type[
        Union[TensorflowBackendInferenceLearner, TFLiteBackendInferenceLearner]
    ],
] = {
    "tf": TensorflowBackendInferenceLearner,
    "tflite": TFLiteBackendInferenceLearner,
}



import os
from typing import Tuple

import torch

from nebullvm.base import ModelParams
from nebullvm.converters.torch_converters import convert_torch_to_onnx

INPUT_SHAPE = (3, 256, 256)
OUTPUT_SHAPE = (2,)
STATIC_BATCH_SIZE = 1
DYNAMIC_BATCH_SIZE = 2


class TestModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = torch.nn.Conv2d(
            in_channels=3, out_channels=64, kernel_size=3
        )
        self.relu1 = torch.nn.ReLU()
        self.conv2 = torch.nn.Conv2d(
            in_channels=64, out_channels=32, kernel_size=3
        )
        self.relu2 = torch.nn.ReLU()
        self.fcn = torch.nn.Linear(32, 2)

    def forward(self, input_tensor_0, input_tensor_1):
        x0 = self.relu2(self.conv2(self.relu1(self.conv1(input_tensor_0))))
        x1 = self.relu2(self.conv2(self.relu1(self.conv1(input_tensor_1))))
        x = x0 + x1
        x = self.fcn(x.mean(dim=(-2, -1)).view(-1, 32))
        return x


def _build_static_model() -> Tuple[torch.nn.Module, ModelParams]:
    model_params = {
        "batch_size": STATIC_BATCH_SIZE,
        "input_infos": [
            {"size": INPUT_SHAPE, "dtype": "float"},
            {"size": INPUT_SHAPE, "dtype": "float"},
        ],
        "output_sizes": [OUTPUT_SHAPE],
    }
    model_params = ModelParams(**model_params)
    model = TestModel()
    return model, model_params


def _build_dynamic_model() -> Tuple[torch.nn.Module, ModelParams]:
    model = TestModel()
    model_params = {
        "batch_size": DYNAMIC_BATCH_SIZE,
        "input_infos": [
            {"size": INPUT_SHAPE, "dtype": "float"},
            {"size": INPUT_SHAPE, "dtype": "float"},
        ],
        "output_sizes": [OUTPUT_SHAPE],
        "dynamic_info": {
            "inputs": [{0: "batch_size"}, {0: "batch_size"}],
            "outputs": [{0: "batch_size"}],
        },
    }
    return model, ModelParams(**model_params)


def get_onnx_model(temp_dir: str, dynamic: bool = False):
    model_path = os.path.join(temp_dir, "test_model.onnx")
    if dynamic:
        model, model_params = _build_dynamic_model()
    else:
        model, model_params = _build_static_model()
    convert_torch_to_onnx(model, model_params, model_path)
    return model_path, model_params



# flake8: noqa

from nebullvm.converters.converters import BaseConverter, ONNXConverter
from nebullvm.converters.tensorflow_converters import (
    convert_tf_to_onnx,
    convert_keras_to_onnx,
)
from nebullvm.converters.torch_converters import convert_torch_to_onnx

__all__ = [k for k in globals().keys() if not k.startswith("_")]



from pathlib import Path
import subprocess
from tempfile import TemporaryDirectory
from typing import Union

import tensorflow as tf
import tf2onnx

from nebullvm.base import ModelParams, DataType
from nebullvm.config import ONNX_OPSET_VERSION


def convert_tf_to_onnx(
    model: tf.Module,
    output_file_path: Union[str, Path],
):
    """Convert TF models into ONNX.

    Args:
        model (tf.Module): TF model.
        output_file_path (Path): Path where storing the output file.
    """
    with TemporaryDirectory() as temp_dir:
        tf.saved_model.save(model, export_dir=temp_dir)
        onnx_cmd = [
            "python3",
            "-m",
            "tf2onnx.convert",
            "--saved-model",
            f"{temp_dir}",
            "--output",
            f"{output_file_path}",
            "--opset",
            f"{ONNX_OPSET_VERSION}",
        ]
        subprocess.run(onnx_cmd)


def convert_keras_to_onnx(
    model: tf.keras.Model,
    model_params: ModelParams,
    output_file_path: Union[str, Path],
):
    """Convert keras models into ONNX.

    Args:
        model (tf.Module): keras model.
        model_params (ModelParams): Model Parameters as input sizes and
            dynamic axis information.
        output_file_path (Path): Path where storing the output file.
    """
    spec = (
        tf.TensorSpec(
            (model_params.batch_size, *input_info.size),
            tf.float32 if input_info.dtype is DataType.FLOAT else tf.int32,
            name=f"input_{i}",
        )
        for i, input_info in enumerate(model_params.input_infos)
    )
    tf2onnx.convert.from_keras(
        model,
        input_signature=spec,
        opset=ONNX_OPSET_VERSION,
        output_path=output_file_path,
    )



import copy
from abc import ABC, abstractmethod
from typing import List, Any, Dict


class BaseTransformation(ABC):
    @abstractmethod
    def _transform(self, _input: Any, **kwargs) -> Any:
        raise NotImplementedError()

    def __call__(self, _input: Any, **kwargs):
        return self._transform(_input, **kwargs)

    def to_dict(self):
        return {
            "module": self.__class__.__module__,
            "name": self.__class__.__name__,
        }

    @classmethod
    def from_dict(cls, tfm_dict: Dict):
        return cls()


class MultiStageTransformation(BaseTransformation):
    def __init__(self, transformations: List[BaseTransformation]):
        self._tfms = transformations

    def _transform(self, _input: Any, **kwargs) -> Any:
        for tfm in self._tfms:
            _input = tfm(_input, **kwargs)
        return _input

    def append(self, __tfm: BaseTransformation):
        self._tfms.append(__tfm)

    def extend(self, tfms: List[BaseTransformation]):
        self._tfms += tfms

    def to_dict(self) -> Dict:
        return {"tfms": [tfm.to_dict() for tfm in self._tfms]}

    @classmethod
    def from_dict(cls, tfms_dict: Dict):
        tfms = []
        for tfm_dict in tfms_dict["tfms"]:
            exec(f"from {tfm_dict['module']} import {tfm_dict['name']}")
            tfm = eval(tfm_dict["name"]).from_dict(tfm_dict)
            tfms.append(tfm)
        return cls(tfms)

    def copy(self):
        new_list = copy.deepcopy(self._tfms)
        return self.__class__(new_list)

    def __len__(self):
        return len(self._tfms)


class NoOp(BaseTransformation):
    def _transform(self, _input: Any, **kwargs):
        return _input



from tempfile import TemporaryDirectory

import pytest

from nebullvm.base import DeepLearningFramework
from nebullvm.inference_learners.tvm import TVM_INFERENCE_LEARNERS
from nebullvm.optimizers import ApacheTVMOptimizer
from nebullvm.optimizers.tests.utils import get_onnx_model


@pytest.mark.parametrize(
    ("output_library", "dynamic"),
    [
        (DeepLearningFramework.PYTORCH, True),
        (DeepLearningFramework.PYTORCH, False),
    ],
)
def test_tvm(output_library: DeepLearningFramework, dynamic: bool):
    with TemporaryDirectory() as tmp_dir:
        model_path, model_params = get_onnx_model(tmp_dir, dynamic)
        optimizer = ApacheTVMOptimizer()
        model = optimizer.optimize(model_path, output_library, model_params)
        assert isinstance(model, TVM_INFERENCE_LEARNERS[output_library])

        inputs_example = model.get_inputs_example()
        res = model.predict(*inputs_example)
        assert res is not None

        if dynamic:
            inputs_example = [
                input_[: len(input_) // 2] for input_ in inputs_example
            ]
            res = model.predict(*inputs_example)
            assert res is not None



from typing import Any, List, Tuple, Dict

from nebullvm.measure import compute_relative_difference, compute_accuracy_drop

QUANTIZATION_METRIC_MAP = {
    "accuracy": compute_accuracy_drop,
    "precision": compute_relative_difference,
}


def check_inputs(
    input_data: Any, batch_size: int, input_sizes: List[Tuple[int, ...]]
):
    if input_data is None and (batch_size is None or input_sizes is None):
        raise ValueError(
            "Non-admissible input detected. You need to specify either the "
            "input data or directly the batch size and the input sizes."
        )


def ifnone(target, new_value):
    if target is None:
        return new_value
    else:
        return target


def inspect_dynamic_size(
    tensors: Tuple[Any, ...],
    sizes: List[Tuple[int, ...]],
    batch_size: int,
    axis_list: List[Dict],
):
    for idx, (tensor, size) in enumerate(zip(tensors, sizes)):
        size = (batch_size, *size)
        for idy, (j, k) in enumerate(zip(tensor.shape, size)):
            if j != k:
                if idy == 0:
                    tag = "batch_size"
                else:
                    tag = f"val_{j}_{k}"
                axis_list[idx][idy] = tag



from tempfile import TemporaryDirectory
from typing import Callable, Optional

import tensorflow as tf

from nebullvm.base import DeepLearningFramework, ModelParams, QuantizationType
from nebullvm.inference_learners.tensorflow import (
    TensorflowBackendInferenceLearner,
    TF_BACKEND_LEARNERS_DICT,
)
from nebullvm.optimizers import BaseOptimizer
from nebullvm.optimizers.quantization.tensorflow import quantize_tf
from nebullvm.optimizers.quantization.utils import (
    check_quantization,
    check_precision,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import convert_to_target_framework
from nebullvm.utils.tf import create_model_inputs_tf, run_tf_model


class TensorflowBackendOptimizer(BaseOptimizer):
    """Optimizer working directly on the tensorflow backend, with no need of a
    conversion to ONNX. The model will be finally compiled using tflite.
    For avoiding un-wanted modification to the input model models are copied
    before being optimized.

    Attributes:
        logger (Logger, optional): Optional logger for logging optimization
            information.
    """

    def optimize(
        self,
        model: tf.Module,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[TensorflowBackendInferenceLearner]:
        """Optimize the input model using pytorch built-in techniques.

        Args:
            model (tf.Module): The tensorflow model. For avoiding un-wanted
                modifications to the original model, it will be copied in the
                method.
            output_library (DeepLearningFramework): Output framework. At the
                current stage just TENSORFLOW is supported.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            TensorflowBackendInferenceLearner or TFLiteBackendInferenceLearner:
                Model optimized for inference.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        assert output_library is DeepLearningFramework.TENSORFLOW, (
            "Other APIs than the Tensorflow one are not supported "
            "for the Tensorflow Backend yet."
        )

        check_quantization(quantization_type, perf_loss_ths)
        with TemporaryDirectory() as tmp_dir:
            if perf_loss_ths is not None:
                if input_data is None:
                    input_data_tf = [
                        tuple(
                            create_model_inputs_tf(
                                model_params.batch_size,
                                model_params.input_infos,
                            )
                        )
                    ]
                    ys = None
                else:
                    input_data_tf, ys = input_data.get_numpy_list(
                        300, with_ys=True
                    )
                    input_data_tf = [
                        tuple(
                            convert_to_target_framework(t, output_library)
                            for t in data_tuple
                        )
                        for data_tuple in input_data_tf
                    ]
                output_data_tf = [
                    tuple(run_tf_model(model, input_tensors))
                    for input_tensors in input_data_tf
                ]
                model, input_tfms = quantize_tf(
                    model=model,
                    quantization_type=quantization_type,
                    input_tfms=input_tfms,
                    input_data=input_data_tf,
                    tmp_dir=tmp_dir,
                )

            learner = TF_BACKEND_LEARNERS_DICT[
                "tflite" if perf_loss_ths is not None else "tf"
            ](model, network_parameters=model_params, input_tfms=input_tfms)
            if perf_loss_ths is not None:
                is_valid = check_precision(
                    learner,
                    input_data_tf,
                    output_data_tf,
                    perf_loss_ths,
                    metric_func=perf_metric,
                    ys=ys,
                )
                if not is_valid:
                    return None
        return learner



from logging import Logger
from typing import Dict, List, Optional, Callable

from onnxruntime.transformers.optimizer import MODEL_TYPES

from nebullvm.base import ModelParams, DeepLearningFramework, QuantizationType
from nebullvm.inference_learners.onnx import (
    ONNXInferenceLearner,
    ONNX_INFERENCE_LEARNERS,
)
from nebullvm.optimizers import BaseOptimizer
from nebullvm.optimizers.quantization.utils import check_precision
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    get_output_names,
    convert_to_numpy,
    run_onnx_model,
)

try:
    from onnxruntime.transformers import optimizer
except ImportError:
    import warnings

    warnings.warn(
        "No valid onnxruntime installation found. Trying to install it..."
    )
    from nebullvm.installers.installers import install_onnxruntime

    install_onnxruntime()
    from onnxruntime.transformers import optimizer


class HuggingFaceOptimizer(BaseOptimizer):
    def __init__(
        self,
        hugging_face_params: Dict,
        perf_loss_ths: float = None,
        perf_metric: Callable = None,
        logger: Logger = None,
    ):
        super(HuggingFaceOptimizer, self).__init__(logger)
        self.hf_params = hugging_face_params
        self.perf_loss_ths = perf_loss_ths
        self.perf_metric = perf_metric
        self.q_type = QuantizationType.HALF

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[ONNXInferenceLearner]:
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        optimized_model = optimizer.optimize_model(model, **self.hf_params)
        if perf_loss_ths is not None:
            if quantization_type is not QuantizationType.HALF:
                return None
            optimized_model.convert_float_to_float16()
            new_onnx_model = model.replace(".onnx", "_fp16.onnx")
        else:
            new_onnx_model = model.replace(".onnx", "_opt.onnx")
        optimized_model.save_model_to_file(new_onnx_model)
        learner = ONNX_INFERENCE_LEARNERS[output_library](
            input_tfms=input_tfms,
            network_parameters=model_params,
            onnx_path=new_onnx_model,
            input_names=get_input_names(new_onnx_model),
            output_names=get_output_names(new_onnx_model),
        )
        if perf_loss_ths is not None:
            # TODO: Add dataset and metric from user
            if input_data is None:
                inputs = [learner.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_list(100, with_ys=True)
            inputs_onnx = [
                tuple(convert_to_numpy(x) for x in input_) for input_ in inputs
            ]
            base_outputs = [
                tuple(run_onnx_model(model, list(input_onnx)))
                for input_onnx in inputs_onnx
            ]
            is_valid = check_precision(
                learner,
                inputs,
                base_outputs,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return learner

    @staticmethod
    def get_accepted_types() -> List[str]:
        return list(MODEL_TYPES.keys())



# flake8: noqa

from nebullvm.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    InferenceLearnerWrapper,
)
from nebullvm.inference_learners.onnx import (
    ONNXInferenceLearner,
    PytorchONNXInferenceLearner,
    TensorflowONNXInferenceLearner,
)
from nebullvm.inference_learners.openvino import (
    OpenVinoInferenceLearner,
    PytorchOpenVinoInferenceLearner,
    TensorflowOpenVinoInferenceLearner,
)
from nebullvm.inference_learners.tensor_rt import (
    NvidiaInferenceLearner,
    PytorchNvidiaInferenceLearner,
    TensorflowNvidiaInferenceLearner,
)
from nebullvm.inference_learners.tvm import (
    ApacheTVMInferenceLearner,
    PytorchApacheTVMInferenceLearner,
    TensorflowApacheTVMInferenceLearner,
)

__all__ = [k for k in globals().keys() if not k.startswith("_")]



from typing import Sequence, List, Tuple, Any, Union, Iterable

import numpy as np

from nebullvm.utils.onnx import convert_to_numpy


class DataManager:
    """Class for managing the user data in nebullvm.

    Attributes:
        data_reader(Sequence): Object implementing the __getitem__, the
            __len__ and the __iter__/__next__ APIs. It should read the
            user data and return tuples of tensors for feeding the models.
    """

    def __init__(self, data_reader: Sequence):
        self._data_reader = data_reader
        self._pointer = 0

    def __getitem__(self, item):
        return self._data_reader[item]

    def __len__(self):
        return len(self._data_reader)

    def __iter__(self):
        self._pointer = 0
        return self

    def __next__(self):
        if self._pointer < len(self):
            data = self[self._pointer]
            self._pointer += 1
            return data
        else:
            raise StopIteration

    def get_numpy_list(
        self, n: int, shuffle: bool = False, with_ys: bool = False
    ) -> Union[
        List[Tuple[np.ndarray, ...]], Tuple[List[Tuple[np.ndarray, ...]], List]
    ]:
        if not with_ys:
            return [
                tuple(convert_to_numpy(x) for x in tuple_)
                for tuple_ in self.get_list(n, shuffle)
            ]
        else:
            xs, ys = self.get_list(n, shuffle, with_ys=True)
            return [
                tuple(convert_to_numpy(x) for x in tuple_) for tuple_ in xs
            ], ys

    def get_list(
        self, n: int, shuffle: bool = False, with_ys: bool = False
    ) -> Union[List[Tuple[Any, ...]], Tuple[List[Tuple[Any, ...]], List]]:
        if shuffle:
            idx = np.random.choice(len(self), n, replace=n > len(self))
        else:
            idx = np.arange(0, min(n, len(self)))
            if n > len(self):
                idx = np.concatenate(
                    [
                        idx,
                        np.random.choice(
                            len(self), n - len(self), replace=True
                        ),
                    ]
                )
        if not with_ys:
            return [self[i][0] for i in idx]

        ys, xs = [], []
        for i in idx:
            x, y = self[i]
            xs.append(x)
            ys.append(y)
        return xs, ys

    @classmethod
    def from_iterable(cls, iterable: Iterable, max_length: int = 500):
        return cls([x for i, x in enumerate(iterable) if i < max_length])



from typing import Any

import torch

from nebullvm.transformations.base import BaseTransformation


class VerifyContiguity(BaseTransformation):
    def _transform(self, _input: Any, **kwargs) -> Any:
        if not isinstance(_input, torch.Tensor):
            return _input
        if not _input.is_contiguous():
            _input = _input.contiguous()
        return _input



import os.path
from typing import List, Tuple

import tensorflow as tf

from nebullvm.base import QuantizationType
from nebullvm.config import TENSORFLOW_BACKEND_FILENAMES
from nebullvm.transformations.base import MultiStageTransformation


def _quantize_dynamic(model: tf.Module):
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    tflite_quant_model = converter.convert()
    return tflite_quant_model


def _quantize_static(model: tf.Module, dataset: List[Tuple[tf.Tensor, ...]]):
    def representative_dataset():
        for data_tuple in dataset:
            yield list(data_tuple)

    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    converter.representative_dataset = representative_dataset
    tflite_quant_model = converter.convert()
    return tflite_quant_model


def _half_precision(model: tf.Module):
    converter = tf.lite.TFLiteConverter.from_keras_model(model)
    converter.optimizations = [tf.lite.Optimize.DEFAULT]
    converter.target_spec.supported_types = [tf.float16]
    tflite_quant_model = converter.convert()
    return tflite_quant_model


def quantize_tf(
    model: tf.Module,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data: List[Tuple[tf.Tensor, ...]],
    tmp_dir: str,
):
    if quantization_type is QuantizationType.DYNAMIC:
        quantized_model = _quantize_dynamic(model)
    elif quantization_type is QuantizationType.STATIC:
        quantized_model = _quantize_static(model, input_data)
    elif quantization_type is QuantizationType.HALF:
        quantized_model = _half_precision(model)
    else:
        raise NotImplementedError(
            f"Quantization not supported for type {quantization_type}"
        )

    filepath = os.path.join(
        tmp_dir, TENSORFLOW_BACKEND_FILENAMES["tflite_model"]
    )
    with open(filepath, "wb") as f:
        f.write(quantized_model)
    return filepath, input_tfms



from collections import OrderedDict
from tempfile import TemporaryDirectory
from typing import (
    Tuple,
    Union,
    List,
    Iterable,
    Dict,
    Any,
    Type,
    Callable,
    Optional,
    Sequence,
)

import numpy as np
import torch

from nebullvm import optimize_torch_model
from nebullvm.api.frontend.utils import ifnone, QUANTIZATION_METRIC_MAP
from nebullvm.base import DataType, ModelCompiler
from nebullvm.inference_learners.base import (
    PytorchBaseInferenceLearner,
    InferenceLearnerWrapper,
    LearnerMetadata,
)
from nebullvm.optimizers.extra import HuggingFaceOptimizer

try:
    from transformers import PreTrainedModel, PretrainedConfig
    from transformers.tokenization_utils import PreTrainedTokenizer
except ImportError:
    # add placeholders for function definition
    PreTrainedModel = None
    PreTrainedTokenizer = None


def _flatten_outputs(
    outputs: Union[torch.Tensor, Iterable]
) -> List[torch.Tensor]:
    new_outputs = []
    for output in outputs:
        if isinstance(output, torch.Tensor):
            new_outputs.append(output)
        else:
            flatten_list = _flatten_outputs(output)
            new_outputs.extend(flatten_list)
    return new_outputs


class _TransformerWrapper(torch.nn.Module):
    """Class for wrappering the Transformers and give them an API compatible
    with nebullvm. The class takes and input of the forward method positional
    arguments and transform them in the input dictionaries needed by
    transformers classes. At the end it also flattens their output.
    """

    def __init__(
        self,
        core_model: torch.nn.Module,
        encoded_input: Dict[str, torch.Tensor],
    ):
        super().__init__()
        self.core_model = core_model
        self.inputs_types = OrderedDict()
        for key, value in encoded_input.items():
            self.inputs_types[key] = value.dtype

    def forward(self, *args: torch.Tensor):
        inputs = {
            key: value for key, value in zip(self.inputs_types.keys(), args)
        }
        outputs = self.core_model(**inputs)
        return tuple(_flatten_outputs(outputs.values()))


def _get_size_recursively(
    tensor_tuple: Union[torch.Tensor, Tuple]
) -> List[int]:
    if isinstance(tensor_tuple[0], torch.Tensor):
        return [len(tensor_tuple)]
    else:
        inner_size = _get_size_recursively(tensor_tuple[0])
        return [len(tensor_tuple), *inner_size]


def _get_output_structure(
    text: str,
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizer,
    tokenizer_args: Dict,
) -> Tuple[OrderedDict, Type]:
    """Function needed for saving in a dictionary the output structure of the
    transformers model.
    """
    encoded_input = tokenizer([text], **tokenizer_args)
    output = model(**encoded_input)
    structure = OrderedDict()
    for key, value in output.items():
        if isinstance(value, torch.Tensor):
            structure[key] = None
        else:
            size = _get_size_recursively(value)
            structure[key] = size
    return structure, type(output)


def _restructure_output(
    output: Tuple[torch.Tensor],
    structure: OrderedDict,
    output_type: Any = None,
):
    """Restructure the flatter output using the structure dictionary given as
    input.
    """
    output_dict = {}
    idx = 0
    for key, value in structure.items():
        if value is None:
            output_dict[key] = output[idx]
            idx += 1
        else:
            output_dict[key] = (
                np.array(
                    output[idx : int(np.prod(value)) + idx],  # noqa E203
                    dtype=object,
                )
                .reshape(value)
                .tolist()
            )
            idx += np.prod(value)
    if output_type is not None:
        return output_type(**output_dict)
    return output_dict


class HuggingFaceInferenceLearner(InferenceLearnerWrapper):
    """Class wrapping an InferenceLearner model and giving to it the
    huggingface interface.

    The class fuse both the InterfaceLearner and HuggingFace interfaces, giving
    to the final user a model which can be used whit the prefered API without
    the need of adapting the previous code.

    Attributes:
        network_parameters (ModelParams): Model parameters of the model.
        core_inference_learner (PytorchBaseInferenceLearner): Inference learner
            built using the Pytorch interface.
        output_structure (Dict): Original output structure of the HuggingFace
            model.
        input_names (List[str]): List of all the input keys used for the
            original HuggingFace model.
        output_type (Any, optional): Original output type of the HuggingFace
            model.
    """

    def __init__(
        self,
        core_inference_learner: PytorchBaseInferenceLearner,
        output_structure: OrderedDict,
        input_names: List[str],
        output_type: Any = None,
    ):
        super().__init__(core_inference_learner)
        self.output_structure = output_structure
        self.input_names = input_names
        self.output_type = output_type

    def _save_wrapper_extra_info(self):
        pass

    @staticmethod
    def _load_wrapper_extra_info(builder_inputs: Dict) -> Dict:
        return builder_inputs

    def run(self, *args, **kwargs) -> Any:
        """Run the underlying optimized model for getting a prediction.

        The method has an hybrid interface. It accepts inputs either as
        positional or keyword arguments. If only positional arguments are given
        the method expects the inputs to be in the canonical
        nebullvm interface. If only keyword arguments are given the method
        expects them to be in the HuggingFace interface. Mixed representation
        is not allowed and will result in an error.
        """
        if len(args) > 0 and len(kwargs) > 0:
            raise RuntimeError(
                "Not allowed usage of the predict method. "
                "Either the positional or the keyword arguments must be given."
            )
        if len(args) > 0:
            return self.core_inference_learner(*args)
        inputs = (kwargs.pop(name) for name in self.input_names)
        outputs = self.core_inference_learner(*inputs)
        return _restructure_output(
            outputs, self.output_structure, self.output_type
        )

    def _get_extra_metadata_kwargs(self) -> Dict:
        metadata_kwargs = {
            "output_structure": self.output_structure,
            "output_structure_keys": list(self.output_structure.keys()),
            "input_names": self.input_names,
        }
        if self.output_type is not None:
            metadata_kwargs.update(
                {
                    "output_type": self.output_type.__name__,
                    "output_type_module": self.output_type.__module__,
                }
            )
        return metadata_kwargs

    @staticmethod
    def _convert_metadata_to_inputs(metadata: LearnerMetadata) -> Dict:
        # we need to guarantee the preservation of the output structure
        # elements order.
        output_structure = OrderedDict()
        for key in metadata["output_structure_keys"]:
            output_structure[key] = metadata["output_structure"][key]

        inputs = {
            "output_structure": output_structure,
            "input_names": metadata["input_names"],
        }
        if metadata["output_type"] is not None:
            exec(
                f"from {metadata['output_type_module']} "
                f"import {metadata['output_type']}"
            )
            inputs["output_type"] = eval(metadata["output_type"])
        return inputs


def _get_dynamic_axis(
    text: str,
    tokenizer: PreTrainedTokenizer,
    model: PreTrainedModel,
    tokenizer_args: Dict,
) -> Dict[str, List[Dict[int, str]]]:
    input_1 = tokenizer([text], **tokenizer_args)
    input_2 = tokenizer([text + text], **tokenizer_args)
    input_dicts = []
    for key in input_1.keys():
        input_dict = {}
        for idx, (i, j) in enumerate(
            zip(input_1[key].shape, input_2[key].shape)
        ):
            if i != j:
                input_dict[idx] = f"val_{i}_{j}"
        input_dicts.append(input_dict)

    output_dicts = []
    outputs_1 = _flatten_outputs(model(**input_1).values())
    outputs_2 = _flatten_outputs(model(**input_2).values())
    for o1, o2 in zip(outputs_1, outputs_2):
        output_dict = {}
        for idx, (i, j) in enumerate(zip(o1.shape, o2.shape)):
            if i != j:
                output_dict[idx] = f"val_{i}_{j}"
        output_dicts.append(output_dict)
    return {"inputs": input_dicts, "outputs": output_dicts}


def _extract_input_type(input_value: torch.Tensor):
    if input_value.dtype is torch.float:
        return DataType.FLOAT
    elif input_value.dtype is torch.long:
        return DataType.INT
    else:
        raise NotImplementedError(
            f"Unsupported data format {input_value.dtype}."
        )


def _try_extraction(model_config: PretrainedConfig, keys: List[str]):
    for key in keys:
        if hasattr(model_config, key):
            return getattr(model_config, key)
    return


def _get_extra_optimizer(
    model_config: PretrainedConfig,
) -> List[HuggingFaceOptimizer]:
    config_name = model_config.__class__.__name__.lower()
    for key in HuggingFaceOptimizer.get_accepted_types():
        if key in config_name:
            input_dict = {"model_type": key, "opt_level": 2}
            hidden_dim = _try_extraction(
                model_config, ["n_embd", "d_model", "hidden_size"]
            )
            if hidden_dim is not None:
                input_dict["hidden_size"] = hidden_dim
            n_heads = _try_extraction(
                model_config,
                ["n_head", "num_attention_heads", "encoder_attention_heads"],
            )
            if n_heads is not None:
                input_dict["num_heads"] = n_heads
            return [HuggingFaceOptimizer(hugging_face_params=input_dict)]
    return [HuggingFaceOptimizer(hugging_face_params={})]


class _HFDataset(Sequence):
    def __init__(
        self,
        input_texts: List,
        ys: Optional[List],
        keywords: List[str],
        batch_size: int,
        tokenizer: PreTrainedTokenizer,
        tokenizer_args: Dict,
    ):
        self._input_texts = input_texts
        self._ys = ys
        self._bs = batch_size
        self._keys = keywords
        self._tokenizer = tokenizer
        if self._tokenizer.pad_token is None:
            self._tokenizer.pad_token = self._tokenizer.eos_token
        _tokenizer_args = {"truncation": True, "padding": True}
        _tokenizer_args.update(tokenizer_args)
        self._tokenizer_args = _tokenizer_args

    def __getitem__(self, item: int):
        pointer = self._bs * item
        if pointer >= len(self):
            raise IndexError
        mini_batch = self._input_texts[
            pointer : pointer + self._bs  # noqa E203
        ]
        if self._ys is not None:
            mini_batch_y = self._ys[pointer : pointer + self._bs]  # noqa E203
        else:
            mini_batch_y = None
        encoded_inputs = self._tokenizer(mini_batch, **self._tokenizer_args)
        return tuple(encoded_inputs[key] for key in self._keys), mini_batch_y

    def __len__(self):
        return len(self._input_texts)


def optimize_huggingface_model(
    model: PreTrainedModel,
    tokenizer: PreTrainedTokenizer,
    input_texts: List[str],
    batch_size: int,
    max_input_sizes: List[Tuple[int, ...]],
    save_dir: str,
    extra_input_info: List[Dict] = None,
    use_static_shape: bool = False,
    use_torch_api: bool = False,
    tokenizer_args: Dict = None,
    ignore_compilers: List[str] = None,
    perf_loss_ths: float = None,
    perf_metric: Union[str, Callable] = None,
    ys: List = None,
):
    """Optimize the HuggingFace model.

    This function saves the output model as well in a nebuly-readable format
    in order to avoid temporary-files corruptions which would prevent the model
    saving later in the process.
    Note that TensorRT compiler is currently disabled for Hugginface models
    since in some cases it can cause an untreatable error in the C++ code
    causing the interruption of the optimization.

    Args:
        model (PreTrainedModel): HuggingFace transformers model.
        tokenizer (PreTrainedTokenizer): Tokenizer used for building model's
            inputs.
        input_texts (List[str]): Texts either from the training set or similar
            to the ones contained in the text. If the perf_loss_ths is
            passed the input_text will be used for computing the drop in
            precision and for setting the quantization parameters. If you
            selected a quantization metric needing the input labels you need to
            provide them for each input in the `ys` argument.
        batch_size (int): Batch size needed for the model.
        max_input_sizes (List[Tuple[int]]): List containing the maximum size of
            all the input tensors of the model.
            Note that even just a single tensor is needed as model input,
            this field must be a list containing (in the exposed case)
            a single element). The tuple must contain the maximum value for
            all the input tensor dimensions excluding the batch size.
            This means that the final input tensor size will be considered as
            `(batch_size, *input_tensor_size)`, where `input_tensor_size` is
            one list element of `max_input_sizes`.
        save_dir (str):  Path to the directory where saving the final model.
        extra_input_info (List[Dict], optional): List of extra information
            needed for defining the input tensors, e.g. max_value and min_value
            the tensors can get.
        use_static_shape (bool): Parameter for fixing the accepted input shape.
        use_torch_api (bool): Parameter for using the torch api of compilers
            when available. The actual implementation supports only the torch
            interface for TVM. Note that when running the torch interface
            nebullvm will ignore the ONNX one once the torch implementation
            succeeds. Clearly, in case of failure of the torch API, a second
            tentative will be done with the ONNX interface.
        tokenizer_args (Dict, optional): Extra args needed for the tokenizer.
        ignore_compilers (List[str], optional): List of DL compilers we want
            to ignore while running the optimization. Compiler name should be
            one between "tvm", "tensor RT", "openvino" and "onnxruntime".
        perf_loss_ths (float, optional): Tolerated relative error for
            performing approximation techniques before compiling the model.
            If no value is given, no optimization will be performed. Note that
            it will not be used for compilers using the torch API when
            `use_torch_api` is `True`. Just dynamic quantization will be
            performed, since no data is given as input.
        perf_metric (Union[Callable, str], optional): The metric to
            be used for accepting or refusing a precision-reduction
            optimization proposal. If none is given but a `perf_loss_ths` is
            received, the `nebullvm.measure.compute_relative_difference`
            metric will be used as default one. A user-defined metric can
            be passed as function accepting as inputs two tuples of tensors
            (produced by the baseline and the quantized model) and the related
            original labels.
            For more information see
            `nebullvm.measure.compute_relative_difference` and
            `nebullvm.measure.compute_accuracy_drop`. `perf_metric`
            accepts as value also a string containing the metric name. At the
            current stage the supported metrics are `"precision"` and
            `"accuracy"`.
        ys: List of target labels. For each input in `input_texts` there should
            be the corresponding label. Note that this feature is just used for
            estimating the accuracy drop while running precision-reduction
            techniques. It will be ignored if these techniques are not
            activated.
    """
    if perf_loss_ths is not None and ys is None and perf_metric == "accuracy":
        raise ValueError(
            "You cannot select the accuracy as quantization metric without "
            "providing valid labels!"
        )
    if isinstance(perf_metric, str):
        perf_metric = QUANTIZATION_METRIC_MAP.get(perf_metric)
    tokenizer_args = tokenizer_args or {}
    tokenizer_args.update({"return_tensors": "pt"})
    output_structure, output_type = _get_output_structure(
        text=input_texts[0],
        model=model,
        tokenizer=tokenizer,
        tokenizer_args=tokenizer_args,
    )
    input_example = tokenizer(input_texts[0], **tokenizer_args)
    input_types = [_extract_input_type(v) for v in input_example.values()] or [
        "int"
    ] * len(input_example)
    # The wrapper model is needed for adapt the huggingface transformers API
    # to the one adopted by the nebullvm optimization.
    wrapper_model = _TransformerWrapper(
        core_model=model, encoded_input=input_example
    )
    with TemporaryDirectory() as tmp_dir:
        optimized_model = optimize_torch_model(
            wrapper_model,
            batch_size=batch_size,
            input_sizes=max_input_sizes,
            save_dir=tmp_dir,
            input_types=input_types,
            extra_input_info=extra_input_info,
            use_torch_api=use_torch_api,
            dynamic_axis=_get_dynamic_axis(
                text=input_texts[0],
                tokenizer=tokenizer,
                model=model,
                tokenizer_args=tokenizer_args,
            )
            if not use_static_shape
            else None,
            perf_loss_ths=perf_loss_ths,
            perf_metric=perf_metric,
            dataloader=_HFDataset(
                input_texts,
                ys,
                list(wrapper_model.inputs_types.keys()),
                batch_size,
                tokenizer,
                tokenizer_args,
            ),
            ignore_compilers=list(
                set(
                    (
                        [ModelCompiler.TENSOR_RT.value]
                        if use_static_shape
                        else [
                            ModelCompiler.TENSOR_RT.value,
                            ModelCompiler.APACHE_TVM.value,
                        ]
                    )
                    + ifnone(ignore_compilers, [])
                )
            ),
            custom_optimizers=_get_extra_optimizer(model.config),
        )
        final_model = HuggingFaceInferenceLearner(
            core_inference_learner=optimized_model,
            output_structure=output_structure,
            input_names=list(wrapper_model.inputs_types.keys()),
            output_type=output_type,
        )
        final_model.save(save_dir)

    return final_model.load(save_dir)



from typing import List, Tuple

import torch
from torch.ao.quantization.stubs import QuantStub, DeQuantStub

from nebullvm.base import QuantizationType
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.transformations.precision_tfms import HalfPrecisionTransformation


class _QuantWrapper(torch.nn.Module):
    def __init__(self, model: torch.nn.Module):
        super(_QuantWrapper, self).__init__()
        qconfig = model.qconfig if hasattr(model, "qconfig") else None
        self.quant = QuantStub(qconfig)
        self.model = model
        self.dequant = DeQuantStub()

    def forward(self, *inputs: torch.Tensor):
        inputs = (self.quant(x) for x in inputs)
        outputs = self.model(*inputs)
        return tuple(self.dequant(x) for x in outputs)


def _quantize_dynamic(model: torch.nn.Module):
    layer_types = {
        type(layer)
        for layer in model.children()
        if len(list(layer.parameters())) > 0
    }
    quantized_model = torch.quantization.quantize_dynamic(
        model=model, qconfig_spec=layer_types, dtype=torch.qint8
    )
    return quantized_model


def _quantize_static(
    model: torch.nn.Module, input_data: List[Tuple[torch.Tensor, ...]]
):
    model = _QuantWrapper(model)
    model.qconfig = torch.quantization.get_default_qconfig("fbgemm")
    model = torch.quantization.fuse_modules(model, [["conv", "relu"]])
    model = torch.quantization.prepare(model)
    for tensors in input_data:
        _ = model(*tensors)
    return torch.quantization.convert(model)


def _half_precision(model: torch.nn.Module):
    return model.half()


def quantize_torch(
    model: torch.nn.Module,
    quantization_type: QuantizationType,
    input_tfms: MultiStageTransformation,
    input_data_torch: List[Tuple[torch.Tensor, ...]],
):
    if quantization_type is QuantizationType.HALF:
        input_tfms.append(HalfPrecisionTransformation())
        return _half_precision(model), input_tfms
    elif quantization_type is QuantizationType.STATIC:
        return _quantize_static(model, input_data_torch), input_tfms
    elif quantization_type is QuantizationType.DYNAMIC:
        return _quantize_dynamic(model), input_tfms
    else:
        raise NotImplementedError(
            f"No quantization implemented for quantization "
            f"type {quantization_type}"
        )



from typing import Optional, Callable

from nebullvm.base import ModelParams, DeepLearningFramework, QuantizationType
from nebullvm.inference_learners.onnx import (
    ONNXInferenceLearner,
    ONNX_INFERENCE_LEARNERS,
)
from nebullvm.optimizers import BaseOptimizer
from nebullvm.optimizers.quantization.onnx import quantize_onnx
from nebullvm.optimizers.quantization.utils import (
    check_precision,
    check_quantization,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    get_output_names,
    create_model_inputs_onnx,
    run_onnx_model,
    convert_to_target_framework,
)


class ONNXOptimizer(BaseOptimizer):
    """Class for creating the inference learner running on onnxruntime."""

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[ONNXInferenceLearner]:
        """Build the ONNX runtime learner from the onnx model.

        Args:
            model (str): Path to the saved onnx model.
            output_library (str): DL Framework the optimized model will be
                compatible with.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            ONNXInferenceLearner: Model running on onnxruntime. The model
                will have an interface in the DL library specified in
                `output_library`.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        input_data_onnx, output_data_onnx, ys = [], [], None
        check_quantization(quantization_type, perf_loss_ths)
        if perf_loss_ths is not None:
            if input_data is None:
                input_data_onnx = [
                    tuple(
                        create_model_inputs_onnx(
                            model_params.batch_size, model_params.input_infos
                        )
                    )
                ]
            else:
                input_data_onnx, ys = input_data.get_numpy_list(
                    300, with_ys=True
                )
            output_data_onnx = [
                tuple(run_onnx_model(model, list(input_tensors)))
                for input_tensors in input_data_onnx
            ]
            model, input_tfms = quantize_onnx(
                model, quantization_type, input_tfms, input_data_onnx
            )
        learner = ONNX_INFERENCE_LEARNERS[output_library](
            input_tfms=input_tfms,
            network_parameters=model_params,
            onnx_path=model,
            input_names=get_input_names(model),
            output_names=get_output_names(model),
        )
        if perf_loss_ths is not None:
            inputs = [
                tuple(
                    convert_to_target_framework(t, output_library)
                    for t in data_tuple
                )
                for data_tuple in input_data_onnx
            ]
            is_valid = check_precision(
                learner,
                inputs,
                output_data_onnx,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return learner



import shutil
import warnings
from abc import ABC
from pathlib import Path
from typing import Union, Type, Dict, Any, List, Generator, Tuple, Optional

import numpy as np
import tensorflow as tf
import torch

from nebullvm.config import TVM_FILENAMES, NO_COMPILER_INSTALLATION
from nebullvm.inference_learners.base import (
    BaseInferenceLearner,
    LearnerMetadata,
    PytorchBaseInferenceLearner,
    TensorflowBaseInferenceLearner,
    NumpyBaseInferenceLearner,
)
from nebullvm.base import ModelParams, DeepLearningFramework
from nebullvm.transformations.base import MultiStageTransformation

try:
    import tvm
    from tvm.contrib.graph_executor import GraphModule
    from tvm.runtime import Module
except ImportError:
    # TODO: Remove the False flag for allowing tvm to be installed by
    #  the Auto-Installer.
    if False and not NO_COMPILER_INSTALLATION:
        warnings.warn(
            "Not found any valid tvm installation. "
            "TVM will not be available in the following steps."
        )
        from nebullvm.installers.installers import install_tvm

        install_tvm()
        GraphModule = None
        Module = None
        warnings.warn(
            "TVM has been successfully installed, but it won't be available "
            "until the reset of the python kernel. Please reboot the python "
            "environment for using TVM optimization."
        )
    else:
        warnings.warn(
            "Not found any valid tvm installation. "
            "TVM will not be available in the following steps."
        )
        Module = object
        GraphModule = object


class ApacheTVMInferenceLearner(BaseInferenceLearner, ABC):
    """Model optimized using ApacheTVM.

    The class cannot be directly instantiated, but implements all the core
    methods needed for using ApacheTVM at inference time.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        graph_executor_module (GraphModule): The graph executor. This is the
            central component in the ApacheTVM optimized model execution.
        input_names (List[str]): Names associated to the model input tensors.
        lib (Module): Component needed for loading the ApacheTVM optimized
            model.
        target (str): Target device. It can be wither `llvm` for targeting CPUs
            or "cuda" for targeting GPUs.
        engine_path (Path, optional): Path to the serialized engine. To be used
            after loading the model (avoiding double engine serialization).
    """

    def __init__(
        self,
        graph_executor_module: GraphModule,
        input_names: List[str],
        lib: Module,
        target: str,
        engine_path: Path = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.graph_executor_module = graph_executor_module
        self.input_names = input_names
        self.lib = lib
        self.target = target
        self.engine_path = (
            self._store_file(engine_path)
            if engine_path is not None
            else engine_path
        )

    def _predict_array(
        self, input_arrays: Generator[np.ndarray, None, None]
    ) -> Generator[np.ndarray, None, None]:
        for name, array in zip(self.input_names, input_arrays):
            self.graph_executor_module.set_input(name, array)
        self.graph_executor_module.run()

        tvm_outputs = (
            self.graph_executor_module.get_output(
                i,
                tvm.nd.empty(
                    (
                        self.network_parameters.batch_size,
                        *output_size,
                    )
                ),
            ).numpy()
            for i, output_size in enumerate(
                self.network_parameters.output_sizes
            )
        )
        return tvm_outputs

    def save(self, path: Union[str, Path], **kwargs):
        """Save the model.

        Args:
            path (Path or str): Path to the directory where the model will
                be stored.
            kwargs (Dict): Dictionary of key-value pairs that will be saved in
                the model metadata file.
        """
        path = Path(path)
        metadata = LearnerMetadata.from_model(
            self, input_names=self.input_names, target=self.target, **kwargs
        )
        metadata.save(path)
        if self.engine_path is None:
            self.lib.export_library(path / TVM_FILENAMES["engine"])
        else:
            shutil.copy(self.engine_path, path)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs):
        """Load the model.

        Args:
            path (Path or str): Path to the directory where the model is
                stored.
            kwargs (Dict): Dictionary of additional arguments for the
                `from_runtime_module` class method.

        Returns:
            ApacheTVMInferenceLearner: The optimized model.
        """
        path = Path(path)
        metadata = LearnerMetadata.read(path).to_dict()
        network_parameters = ModelParams(**metadata["network_parameters"])
        lib = tvm.runtime.load_module(path / TVM_FILENAMES["engine"])
        target_device = metadata["target"]
        input_names = metadata["input_names"]
        input_tfms = metadata.get("input_tfms")
        if input_tfms is not None:
            metadata["input_tfms"] = MultiStageTransformation.from_dict(
                input_tfms
            )
        self = cls.from_runtime_module(
            network_parameters=network_parameters,
            lib=lib,
            target_device=target_device,
            input_names=input_names,
        )
        self.engine_path = path / TVM_FILENAMES["engine"]
        return self

    @classmethod
    def from_runtime_module(
        cls,
        network_parameters: ModelParams,
        lib: Module,
        target_device: str,
        input_names: List[str],
        input_tfms: MultiStageTransformation = None,
    ):
        """Build the model from the runtime module (lib).

        Args:
            network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
            lib (Module): Component needed for loading the ApacheTVM optimized
                model.
            target_device (str): The target device. Either `llvm` (CPU)
                or `cuda`.
            input_names (List[str]): Names associated to the model input
                tensors.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
        """
        dev = tvm.device(str(target_device), 0)
        graph_executor_module = GraphModule(lib["default"](dev))
        return cls(
            input_tfms=input_tfms,
            network_parameters=network_parameters,
            graph_executor_module=graph_executor_module,
            input_names=input_names,
            lib=lib,
            target=target_device,
        )


class BaseArrayApacheTVMInferenceLearner(ApacheTVMInferenceLearner, ABC):
    """Base Model that can be used for all array-based
    ApacheTVMInferenceLearners.
    """

    def _inner_predict(
        self,
        input_arrays: Generator[np.ndarray, None, None],
        input_shapes: Optional[List[Tuple[int, ...]]],
    ) -> Generator[np.ndarray, None, None]:
        if self.network_parameters.dynamic_info is not None:
            input_arrays = (
                np.pad(
                    input_array,
                    [
                        (0, abs(x - y))
                        for x, y in zip(
                            input_array.shape,
                            (self.network_parameters.batch_size, *input_size),
                        )
                    ],
                    mode="constant",
                    constant_values=0,
                )
                for input_array, input_size in zip(
                    input_arrays, self.network_parameters.input_sizes
                )
            )

        output_arrays = self._predict_array(input_arrays)
        if self.network_parameters.dynamic_info is not None:
            assert input_shapes is not None
            dynamic_info = self.network_parameters.dynamic_info
            return (
                output_array[
                    tuple(
                        slice(
                            0,
                            None
                            if x not in out_dynamic_dict.keys()
                            else dynamic_info.retrieve_output_dim(
                                input_shapes, j, i, x
                            ),
                        )
                        for i, x in enumerate(output_array.shape)
                    )
                ]
                for j, (output_array, out_dynamic_dict) in enumerate(
                    zip(output_arrays, dynamic_info.outputs)
                )
            )

        return output_arrays


class PytorchApacheTVMInferenceLearner(
    BaseArrayApacheTVMInferenceLearner, PytorchBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a Pytorch interface.

    This class can be used exactly in the same way as a pytorch Module object.
    At prediction time it takes as input pytorch tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        graph_executor_module (GraphModule): The graph executor. This is the
            central component in the ApacheTVM optimized model execution.
        input_names (List[str]): Names associated to the model input tensors.
        lib (Module): Component needed for loading the ApacheTVM optimized
            model.
        target (str): Target device. It can be wither `llvm` for targeting CPUs
            or "cuda" for targeting GPUs.
    """

    def run(self, *input_tensors: torch.Tensor) -> Tuple[torch.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        device = self._convert_device(input_tensors[0].get_device())
        input_arrays = (
            input_tensor.cpu().detach().numpy()
            for input_tensor in input_tensors
        )
        input_shapes = (
            [tuple(input_tensor.shape) for input_tensor in input_tensors]
            if self.network_parameters.dynamic_info is not None
            else None
        )
        output_arrays = self._inner_predict(input_arrays, input_shapes)
        return tuple(
            torch.from_numpy(array).to(device) for array in output_arrays
        )

    @staticmethod
    def _convert_device(device: Any):
        if isinstance(device, int):
            return "cpu"
        return device


class TensorflowApacheTVMInferenceLearner(
    BaseArrayApacheTVMInferenceLearner, TensorflowBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        graph_executor_module (GraphModule): The graph executor. This is the
            central component in the ApacheTVM optimized model execution.
        input_names (List[str]): Names associated to the model input tensors.
        lib (Module): Component needed for loading the ApacheTVM optimized
            model.
        target (str): Target device. It can be wither `llvm` for targeting CPUs
            or "cuda" for targeting GPUs.
    """

    def run(self, *input_tensors: tf.Tensor) -> Tuple[tf.Tensor, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[Tensor]): Input tensors belonging to the same
                batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[Tensor]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (input_tensor.numpy() for input_tensor in input_tensors)
        input_shapes = (
            [tuple(input_tensor.shape) for input_tensor in input_tensors]
            if self.network_parameters.dynamic_info is not None
            else None
        )
        return tuple(
            tf.convert_to_tensor(out)
            for out in self._inner_predict(input_arrays, input_shapes)
        )


class NumpyApacheTVMInferenceLearner(
    BaseArrayApacheTVMInferenceLearner, NumpyBaseInferenceLearner
):
    """Model optimized using ApacheTVM with a tensorflow interface.

    This class can be used exactly in the same way as a tf.Module or
    keras.Model object.
    At prediction time it takes as input tensorflow tensors given as positional
    arguments.

    Attributes:
        network_parameters (ModelParams): The model parameters as batch
                size, input and output sizes.
        graph_executor_module (GraphModule): The graph executor. This is the
            central component in the ApacheTVM optimized model execution.
        input_names (List[str]): Names associated to the model input tensors.
        lib (Module): Component needed for loading the ApacheTVM optimized
            model.
        target (str): Target device. It can be wither `llvm` for targeting CPUs
            or "cuda" for targeting GPUs.
    """

    def run(self, *input_tensors: np.ndarray) -> Tuple[np.ndarray, ...]:
        """Predict on the input tensors.

        Note that the input tensors must be on the same batch. If a sequence
        of tensors is given when the model is expecting a single input tensor
        (with batch size >= 1) an error is raised.

        Args:
            input_tensors (Tuple[ndarray]): Input tensors belonging to the
                same batch. The tensors are expected having dimensions
                (batch_size, dim1, dim2, ...).

        Returns:
            Tuple[ndarray]: Output tensors. Note that the output tensors does
                not correspond to the prediction on the input tensors with a
                1 to 1 mapping. In fact the output tensors are produced as the
                multiple-output of the model given a (multi-) tensor input.
        """
        input_arrays = (input_tensor for input_tensor in input_tensors)
        input_shapes = (
            [tuple(input_tensor.shape) for input_tensor in input_tensors]
            if self.network_parameters.dynamic_info is not None
            else None
        )
        return tuple(self._inner_predict(input_arrays, input_shapes))


TVM_INFERENCE_LEARNERS: Dict[
    DeepLearningFramework, Type[ApacheTVMInferenceLearner]
] = {
    DeepLearningFramework.PYTORCH: PytorchApacheTVMInferenceLearner,
    DeepLearningFramework.TENSORFLOW: TensorflowApacheTVMInferenceLearner,
    DeepLearningFramework.NUMPY: NumpyApacheTVMInferenceLearner,
}



import os
import uuid
from typing import Tuple, Dict, Optional, Callable

import onnx
import torch.cuda

from nebullvm.base import ModelParams, DeepLearningFramework, QuantizationType
from nebullvm.config import (
    AUTO_TVM_TUNING_OPTION,
    AUTO_TVM_PARAMS,
)
from nebullvm.inference_learners.tvm import (
    TVM_INFERENCE_LEARNERS,
    ApacheTVMInferenceLearner,
)
from nebullvm.optimizers.base import BaseOptimizer
from nebullvm.optimizers.quantization.tvm import TVMCalibrator
from nebullvm.optimizers.quantization.utils import (
    check_quantization,
    check_precision,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    create_model_inputs_onnx,
    run_onnx_model,
    convert_to_numpy,
)
from nebullvm.utils.torch import create_model_inputs_torch, run_torch_model

try:
    import tvm
    from tvm import IRModule
    from tvm.runtime.ndarray import NDArray
    from tvm.autotvm.tuner import XGBTuner
    from tvm import autotvm
    import tvm.relay as relay
except ImportError:
    # TVM is installed in the inference_learner package.
    # TVM objects needed for avoiding errors:
    IRModule = object
    NDArray = object


class ApacheTVMOptimizer(BaseOptimizer):
    """Class for compiling the AI models on Nvidia GPUs using TensorRT."""

    def optimize_from_torch(
        self,
        torch_model: torch.nn.Module,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[ApacheTVMInferenceLearner]:
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        target = self._get_target()
        mod, params = self._build_tvm_model_from_torch(
            torch_model, model_params
        )
        if perf_loss_ths is not None:
            if quantization_type is QuantizationType.HALF:
                mod = tvm.relay.transform.ToMixedPrecision(
                    mixed_precision_type="float16"
                )(mod)
            else:
                if quantization_type is QuantizationType.DYNAMIC:
                    inputs = None
                elif quantization_type is QuantizationType.STATIC:
                    if input_data is None:
                        inputs = [
                            tuple(
                                create_model_inputs_onnx(
                                    model_params.batch_size,
                                    model_params.input_infos,
                                )
                            )
                        ]
                    else:
                        inputs = input_data.get_numpy_list(300, with_ys=False)
                else:
                    return
                mod = self._quantize(mod, params, input_data=inputs)
        tuning_records = self._tune_tvm_model(target, mod, params)
        with autotvm.apply_history_best(tuning_records):
            with tvm.transform.PassContext(opt_level=3, config={}):
                lib = relay.build(mod, target=target, params=params)
        model = TVM_INFERENCE_LEARNERS[
            DeepLearningFramework.PYTORCH
        ].from_runtime_module(
            input_tfms=input_tfms,
            network_parameters=model_params,
            lib=lib,
            target_device=target,
            input_names=[
                f"input_{i}" for i in range(len(model_params.input_infos))
            ],
        )
        if quantization_type is not None:
            if input_data is None:
                inputs = [model.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_list(
                    100, shuffle=True, with_ys=True
                )
            output_data = [
                tuple(run_torch_model(torch_model, list(tuple_)))
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                model,
                inputs,
                output_data,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return model

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[ApacheTVMInferenceLearner]:
        """Optimize the input model with Apache TVM.

        Args:
            model (str): Path to the saved onnx model.
            output_library (str): DL Framework the optimized model will be
                compatible with.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            ApacheTVMInferenceLearner: Model optimized with TVM. The model
                will have an interface in the DL library specified in
                `output_library`.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        check_quantization(quantization_type, perf_loss_ths)
        target = self._get_target()
        mod, params = self._build_tvm_model_from_onnx(model, model_params)
        if perf_loss_ths is not None:
            if quantization_type is QuantizationType.HALF:
                mod = tvm.relay.transform.ToMixedPrecision(
                    mixed_precision_type="float16"
                )(mod)
            else:
                if quantization_type is QuantizationType.DYNAMIC:
                    inputs = None
                elif quantization_type is QuantizationType.STATIC:
                    if input_data is None:
                        inputs = [
                            tuple(
                                create_model_inputs_onnx(
                                    model_params.batch_size,
                                    model_params.input_infos,
                                )
                            )
                        ]
                    else:
                        inputs = input_data.get_numpy_list(300, with_ys=False)
                    inputs = TVMCalibrator(inputs, get_input_names(model))
                else:
                    return
                mod = self._quantize(mod, params, input_data=inputs)
        tuning_records = self._tune_tvm_model(target, mod, params)
        with autotvm.apply_history_best(tuning_records):
            with tvm.transform.PassContext(opt_level=3, config={}):
                lib = relay.build(mod, target=target, params=params)
        model = TVM_INFERENCE_LEARNERS[output_library].from_runtime_module(
            input_tfms=input_tfms,
            network_parameters=model_params,
            lib=lib,
            target_device=target,
            input_names=get_input_names(model),
        )
        if quantization_type is not None:
            if input_data is None:
                inputs = [model.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_list(
                    100, shuffle=True, with_ys=True
                )
            output_data = [
                tuple(
                    run_onnx_model(
                        model,
                        [convert_to_numpy(x) for x in tuple_],
                    )
                )
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                model,
                inputs,
                output_data,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return model

    @staticmethod
    def _build_tvm_model_from_torch(
        torch_model: torch.nn.Module, model_params: ModelParams
    ) -> Tuple[IRModule, Dict[str, NDArray]]:
        shape_dict = {
            f"input_{i}": (
                model_params.batch_size,
                *input_size,
            )
            for i, input_size in enumerate(model_params.input_sizes)
        }
        inputs = tuple(
            create_model_inputs_torch(
                model_params.batch_size, model_params.input_infos
            )
        )
        if torch.cuda.is_available():
            inputs = tuple(input_.cpu() for input_ in inputs)
            torch_model.cpu()
        with torch.no_grad():
            _ = torch_model(*inputs)
            model_trace = torch.jit.trace(torch_model, inputs)
            model_trace.eval()
        mod, params = relay.frontend.from_pytorch(
            model_trace, list(shape_dict.items())
        )
        return mod, params

    @staticmethod
    def _build_tvm_model_from_onnx(
        onnx_model_path: str, model_params: ModelParams
    ) -> Tuple[IRModule, Dict[str, NDArray]]:
        shape_dict = {
            input_key: (
                model_params.batch_size,
                *input_size,
            )
            for input_key, input_size in zip(
                get_input_names(onnx_model_path), model_params.input_sizes
            )
        }
        onnx_model = onnx.load(onnx_model_path)
        mod, params = relay.frontend.from_onnx(onnx_model, shape_dict)
        return mod, params

    @staticmethod
    def _quantize(
        mod: IRModule,
        params: Dict[str, NDArray],
        input_data: TVMCalibrator = None,
    ) -> IRModule:
        if input_data is not None:
            with relay.quantize.qconfig(
                calibrate_mode="kl_divergence", weight_scale="max"
            ):
                mod = relay.quantize.quantize(mod, params, dataset=input_data)
        else:
            with relay.quantize.qconfig(
                calibrate_mode="global_scale", global_scale=8.0
            ):
                mod = relay.quantize.quantize(mod, params)
        return mod

    @staticmethod
    def _get_target() -> str:
        force_on_cpu = int(os.getenv("TVM_ON_CPU", 0)) > 1
        if not force_on_cpu and torch.cuda.is_available():
            return str(tvm.target.cuda())
        else:
            return "llvm"  # run on CPU

    @staticmethod
    def _tune_tvm_model(
        target: str, mod: IRModule, params: Dict[str, NDArray]
    ) -> str:
        """Tune the model using AutoTVM."""
        # TODO: add support to Ansor
        tuning_records = f"{uuid.uuid4()}_model_records.json"
        # create a TVM runner
        runner = autotvm.LocalRunner(
            number=AUTO_TVM_PARAMS["number"],
            repeat=AUTO_TVM_PARAMS["repeat"],
            timeout=AUTO_TVM_PARAMS["timeout"],
            min_repeat_ms=AUTO_TVM_PARAMS["min_repeat_ms"],
            # TODO modify min_repeat_ms for GPU usage
            enable_cpu_cache_flush=True,
        )
        # begin by extracting the tasks from the onnx model
        tasks = autotvm.task.extract_from_program(
            mod["main"], target=target, params=params
        )

        # Tune the extracted tasks sequentially.
        for i, task in enumerate(tasks):
            tuner_obj = XGBTuner(task, loss_type="rank")
            tuner_obj.tune(
                n_trial=min(
                    AUTO_TVM_TUNING_OPTION["trials"], len(task.config_space)
                ),
                early_stopping=AUTO_TVM_TUNING_OPTION["early_stopping"],
                measure_option=autotvm.measure_option(
                    builder=autotvm.LocalBuilder(build_func="default"),
                    runner=runner,
                ),
                callbacks=[
                    autotvm.callback.log_to_file(tuning_records),
                ],
            )
        return tuning_records



from pathlib import Path
from typing import List, Tuple, Optional, Callable

import numpy as np
import torch

from nebullvm.base import DeepLearningFramework, ModelParams, QuantizationType
from nebullvm.config import NVIDIA_FILENAMES, NO_COMPILER_INSTALLATION
from nebullvm.inference_learners.tensor_rt import (
    NVIDIA_INFERENCE_LEARNERS,
    NvidiaInferenceLearner,
)
from nebullvm.optimizers.base import (
    BaseOptimizer,
)
from nebullvm.optimizers.quantization.tensor_rt import TensorRTCalibrator
from nebullvm.optimizers.quantization.utils import (
    check_precision,
    check_quantization,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    get_output_names,
    create_model_inputs_onnx,
    run_onnx_model,
    convert_to_numpy,
)

if torch.cuda.is_available():
    try:
        import tensorrt as trt
    except ImportError:
        from nebullvm.installers.installers import install_tensor_rt
        import warnings

        if not NO_COMPILER_INSTALLATION:
            warnings.warn(
                "No TensorRT valid installation has been found. "
                "Trying to install it from source."
            )
            install_tensor_rt()
            import tensorrt as trt
        else:
            warnings.warn(
                "No TensorRT valid installation has been found. "
                "It won't be possible to use it in the following."
            )


class TensorRTOptimizer(BaseOptimizer):
    """Class for compiling the AI models on Nvidia GPUs using TensorRT."""

    def _build_and_save_the_engine(
        self,
        engine_path: str,
        onnx_model_path: str,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation,
        quantization_type: QuantizationType = None,
        input_data: List[Tuple[np.ndarray, ...]] = None,
    ):
        # -- Build phase --
        nvidia_logger = trt.Logger(trt.Logger.WARNING)
        builder = trt.Builder(nvidia_logger)
        # create network definition
        network = builder.create_network(
            1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
        )
        # build the engine
        # TODO: setup config value for the class in a config file
        config = builder.create_builder_config()
        config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, 1 << 30)
        if quantization_type is QuantizationType.HALF:
            config.set_flag(trt.BuilderFlag.FP16)
        elif quantization_type is QuantizationType.STATIC:
            assert input_data is not None, (
                "You need to specify the calibration data for "
                "performing static quantization."
            )
            calibrator = TensorRTCalibrator(
                batch_size=model_params.batch_size,
                input_data=input_data,
            )
            config.set_flag(trt.BuilderFlag.INT8)
            config.int8_calibrator = calibrator
        elif quantization_type is QuantizationType.DYNAMIC:
            # onnx_model_path, _ = quantize_onnx(
            #     onnx_model_path, quantization_type, input_tfms, input_data
            # )
            config.set_flag(trt.BuilderFlag.INT8)
        # import the model
        parser = trt.OnnxParser(network, nvidia_logger)
        success = parser.parse_from_file(onnx_model_path)

        if not success:
            for idx in range(parser.num_errors):
                if self.logger is not None:
                    self.logger.debug(parser.get_error(idx))
            raise ValueError(
                f"Errors occurred while processing the "
                f"ONNX file at {onnx_model_path}"
            )

        if model_params.dynamic_info is not None:
            profile = builder.create_optimization_profile()
            for input_name, input_dynamic_info, input_info in zip(
                get_input_names(onnx_model_path),
                model_params.dynamic_info.inputs,
                model_params.input_infos,
            ):
                profile.set_shape(
                    input_name,
                    (
                        min(model_params.batch_size, 1)
                        if 0 in input_dynamic_info
                        else model_params.batch_size,
                        *(
                            shape
                            if i + 1 not in input_dynamic_info
                            else (input_info.min_sizes or {}).get(i + 1, 1)
                            for i, shape in enumerate(input_info.size)
                        ),
                    ),
                    (model_params.batch_size, *input_info.size),
                    (model_params.batch_size, *input_info.size),
                )
            config.add_optimization_profile(profile)
        serialized_engine = builder.build_serialized_network(network, config)
        with open(engine_path, "wb") as f:
            f.write(serialized_engine)

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[NvidiaInferenceLearner]:
        """Optimize the input model with TensorRT.

        Args:
            model (str): Path to the saved onnx model.
            output_library (str): DL Framework the optimized model will be
                compatible with.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            TensorRTInferenceLearner: Model optimized with TensorRT. The model
                will have an interface in the DL library specified in
                `output_library`.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        if not torch.cuda.is_available():
            raise SystemError(
                "You are trying to run an optimizer developed for NVidia gpus "
                "on a machine not connected to any GPU supporting CUDA."
            )
        check_quantization(quantization_type, perf_loss_ths)
        engine_path = Path(model).parent / NVIDIA_FILENAMES["engine"]
        if (
            perf_loss_ths is not None
            and quantization_type is QuantizationType.STATIC
        ):
            if input_data is None:
                input_data_onnx = [
                    tuple(
                        create_model_inputs_onnx(
                            model_params.batch_size, model_params.input_infos
                        )
                    )
                ]
            else:
                input_data_onnx = input_data.get_numpy_list(300, with_ys=False)
        elif (
            perf_loss_ths is not None
            and quantization_type is QuantizationType.DYNAMIC
        ):
            return None  # Dynamic quantization is not supported on tensorRT
        else:
            input_data_onnx = None
        self._build_and_save_the_engine(
            engine_path=engine_path,
            onnx_model_path=model,
            model_params=model_params,
            input_tfms=input_tfms,
            quantization_type=quantization_type,
            input_data=input_data_onnx,
        )
        learner = NVIDIA_INFERENCE_LEARNERS[output_library].from_engine_path(
            input_tfms=input_tfms,
            network_parameters=model_params,
            engine_path=engine_path,
            input_names=get_input_names(model),
            output_names=get_output_names(model),
        )
        if quantization_type is not None:
            if input_data is None:
                inputs = [learner.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_numpy_list(
                    100, shuffle=True, with_ys=True
                )
            output_data = [
                tuple(
                    run_onnx_model(
                        model,
                        [convert_to_numpy(x) for x in tuple_],
                    )
                )
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                learner,
                inputs,
                output_data,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return learner



from typing import List, Sequence

from nebullvm.utils.data import DataManager

try:
    from tvm.runtime.ndarray import NDArray
except ImportError:
    NDArray = object


class TVMCalibrator(DataManager):
    def __init__(self, data_reader: Sequence, input_names: List[str]):
        super(TVMCalibrator, self).__init__(data_reader=data_reader)
        self._input_names = input_names

    def __getitem__(self, item: int):
        tuple_ = self._data_reader[item]
        return {name: data for name, data in zip(self._input_names, tuple_)}



from nebullvm.api.frontend.torch import optimize_torch_model  # noqa F401
from nebullvm.api.frontend.tf import optimize_tf_model  # noqa F401
from nebullvm.api.frontend.onnx import optimize_onnx_model  # noqa F401

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import os
import shutil
from tempfile import TemporaryDirectory
from typing import List, Tuple, Dict, Optional, Callable, Union

import numpy as np

from nebullvm.api.frontend.utils import (
    inspect_dynamic_size,
    ifnone,
    QUANTIZATION_METRIC_MAP,
)
from nebullvm.base import (
    InputInfo,
    DeepLearningFramework,
    ModelParams,
    ModelCompiler,
)
from nebullvm.inference_learners.base import NumpyBaseInferenceLearner
from nebullvm.optimizers import BaseOptimizer
from nebullvm.optimizers.multi_compiler import MultiCompilerOptimizer
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    create_model_inputs_onnx,
    get_output_sizes_onnx,
    run_onnx_model,
)


def _extract_dynamic_axis(
    onnx_model: str,
    data: List[Tuple[Tuple[np.ndarray, ...], np.ndarray]],
    input_sizes: List[Tuple[int, ...]],
    batch_size: int,
    max_data: int = 100,
) -> Optional[Dict]:
    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, (input_tensors, y) in enumerate(data):
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, batch_size, dynamic_axis["inputs"]
        )
        outputs = tuple(run_onnx_model(onnx_model, list(input_tensors)))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(
            outputs, output_sizes, batch_size, dynamic_axis["outputs"]
        )
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def _extract_info_from_data(
    onnx_model: str,
    data: List[Tuple[Tuple[np.ndarray, ...], np.ndarray]],
    batch_size: int,
    input_sizes: List[Tuple[int, ...]],
    input_types: List[str],
    dynamic_axis: Dict,
):
    input_row, _ = data[0]
    batch_size = ifnone(batch_size, int(input_row[0].shape[0]))
    input_sizes = ifnone(input_sizes, [tuple(x.shape[1:]) for x in input_row])
    input_types = ifnone(
        input_types, ["int" if x.dtype == int else "float" for x in input_row]
    )
    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(onnx_model, data, input_sizes, batch_size),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def optimize_onnx_model(
    model_path: str,
    save_dir: str,
    data: List[Tuple[Tuple[np.ndarray, ...], np.ndarray]] = None,
    batch_size: int = None,
    input_sizes: List[Tuple[int, ...]] = None,
    input_types: List[str] = None,
    extra_input_info: List[Dict] = None,
    dynamic_axis: Dict = None,
    perf_loss_ths: float = None,
    perf_metric: Union[str, Callable] = None,
    ignore_compilers: List[str] = None,
    custom_optimizers: List[BaseOptimizer] = None,
) -> NumpyBaseInferenceLearner:
    """Basic function for optimizing an onnx model.

    This function saves the output model as well in a nebuly-readable format
    in order to avoid temporary-files corruptions which would prevent the model
    saving later in the process.

    Args:
        model_path (str): ONNX model that needs optimization.
        save_dir (str): Path to the directory where saving the final model.
        data (List):  List of tuples in the form of (xs, y) where xs are
            tuples of np.ndarray and ys can be whatever needed for computing
            the selected metric at quantization time. The data will be used
            for extracting all the data related information or completing
            the missing information in the case of partially given ones. If no
            data is given, both the `batch_size` and the `input_sizes` must be
            passed by the user.
        batch_size (int, optional): The model batch size.
        input_sizes (List[Tuple]], optional): List containing the size of all
            the input tensors of the model. Note that even just a single
            tensor is needed as model input, this field must be a list
            containing (in the exposed case) a single element). The tuple must
            contain all the input tensor dimensions excluding the batch size.
            This means that the final input tensor size will be considered as
            `(batch_size, *input_tensor_size)`, where `input_tensor_size` is
            one list element of `input_sizes`.
        input_types (List[str], optional): List of input types. If no value is
            given all the inputs will be considered as float type. The
            supported string values are "int" and "float".
        extra_input_info (List[Dict], optional): List of extra information
            needed for defining the input tensors, e.g. max_value and min_value
            the tensors can get.
        dynamic_axis (Dict, optional): Dictionary containing info about the
            dynamic axis. It should contain as keys both "inputs" and "outputs"
            and as values two lists of dictionaries where each dictionary
            represents the dynamic axis information for an input/output tensor.
            The inner dictionary should have as key an integer, i.e. the
            dynamic axis (considering also the batch size) and as value a
            string giving a "tag" to it, e.g. "batch_size".
        perf_loss_ths (float, optional): Tolerated relative error for
            performing approximation techniques before compiling the model.
            If no value is given, no optimization will be performed. Note that
            it will not be used for compilers using the torch API when
            `use_torch_api` is `True`. Just dynamic quantization will be
            performed, since no data is given as input.
        perf_metric (Union[Callable, str], optional): The metric to
            be used for accepting or refusing a precision-reduction
            optimization proposal. If none is given but a `perf_loss_ths` is
            received, the `nebullvm.measure.compute_relative_difference`
            metric will be used as default one. A user-defined metric can
            be passed as function accepting as inputs two tuples of tensors
            (produced by the baseline and the quantized model) and the related
            original labels.
            For more information see
            `nebullvm.measure.compute_relative_difference` and
            `nebullvm.measure.compute_accuracy_drop`. `perf_metric`
            accepts as value also a string containing the metric name. At the
            current stage the supported metrics are `"precision"` and
            `"accuracy"`.
        ignore_compilers (List[str], optional): List of DL compilers we want
            to ignore while running the optimization. Compiler name should be
            one between "tvm", "tensor RT", "openvino" and "onnxruntime".
        custom_optimizers (List[BaseOptimizer], optional): List of optimizers
            which can be used for producing InferenceLearners, i.e. models
            optimized for inference. This list is useful when some compilers,
            not included in the base version of nebullvm or not used
            by default, which are specific for a particular tasks, need
            to be used.

    Returns:
        PytorchBaseInferenceLearner: Optimized model usable with the classical
            Pytorch interface. Note that as a torch model it takes as input
            and it gives as output `torch.Tensor`s.
    """
    if data is not None:
        (
            batch_size,
            input_sizes,
            input_types,
            dynamic_axis,
        ) = _extract_info_from_data(
            model_path,
            data,
            batch_size,
            input_sizes,
            input_types,
            dynamic_axis,
        )
        input_data = DataManager(data)
    else:
        input_data = None
    if isinstance(perf_metric, str):
        perf_metric = QUANTIZATION_METRIC_MAP.get(perf_metric)
    if input_types is None:
        input_types = ["float"] * len(input_sizes)
    if extra_input_info is None:
        extra_input_info = [{}] * len(input_sizes)
    if not len(input_sizes) == len(input_types) == len(extra_input_info):
        raise ValueError(
            f"Mismatch in the input list lengths. Given {len(input_sizes)} "
            f"sizes, {len(input_types)} input types and "
            f"{len(extra_input_info)} extra input infos."
        )
    input_infos = [
        InputInfo(size=input_size, dtype=input_type, **extra_info)
        for input_size, input_type, extra_info in zip(
            input_sizes, input_types, extra_input_info
        )
    ]
    dl_library = DeepLearningFramework.NUMPY
    model_params = ModelParams(
        batch_size=batch_size,
        input_infos=input_infos,
        output_sizes=get_output_sizes_onnx(
            model_path,
            input_tensors=create_model_inputs_onnx(batch_size, input_infos),
        ),
        dynamic_info=dynamic_axis,
    )
    ignore_compilers = (
        []
        if ignore_compilers is None
        else [ModelCompiler(compiler) for compiler in ignore_compilers]
    )
    input_tfms = MultiStageTransformation([])
    with TemporaryDirectory() as tmp_dir:
        onnx_path = shutil.copy(model_path, tmp_dir)
        model_optimizer = MultiCompilerOptimizer(
            ignore_compilers=ignore_compilers,
            extra_optimizers=custom_optimizers,
            debug_mode=int(os.environ.get("DEBUG_MODE", "0")) > 0,
        )
        if model_optimizer.usable:
            model_optimized = model_optimizer.optimize(
                model=str(onnx_path),
                output_library=dl_library,
                model_params=model_params,
                input_tfms=input_tfms,
                perf_loss_ths=perf_loss_ths,
                perf_metric=perf_metric,
                input_data=input_data,
            )
        else:
            model_optimized = None
        if model_optimized is None:
            raise RuntimeError(
                "No valid compiled model has been produced. "
                "Look at the logs for further information about the failure."
            )
        model_optimized.save(save_dir)
    return model_optimized.load(save_dir)



from pathlib import Path
from typing import Union

import torch
from torch.nn import Module

from nebullvm.base import ModelParams
from nebullvm.config import ONNX_OPSET_VERSION
from nebullvm.utils.torch import (
    get_outputs_sizes_torch,
    create_model_inputs_torch,
)


def convert_torch_to_onnx(
    torch_model: Module,
    model_params: ModelParams,
    output_file_path: Union[str, Path],
):
    """Function importing a custom model in pytorch and converting it in ONNX

    Args:
        torch_model (Module): Pytorch model.
        model_params (ModelParams): Model Parameters as input sizes and
            dynamic axis information.
        output_file_path (str or Path): Path where storing the output
            ONNX file.
    """
    input_tensors = create_model_inputs_torch(
        model_params.batch_size, model_params.input_infos
    )
    output_sizes = get_outputs_sizes_torch(torch_model, input_tensors)
    if torch.cuda.is_available():  # move back tensors to cpu for conversion
        input_tensors = [x.cpu() for x in input_tensors]
        torch_model.cpu()
    input_names = [f"input_{i}" for i in range(len(input_tensors))]
    output_names = [f"output_{i}" for i in range(len(output_sizes))]
    dynamic_info = model_params.dynamic_info
    if dynamic_info is not None:
        dynamic_info = {
            name: dynamic_dict
            for name, dynamic_dict in zip(
                input_names + output_names,
                dynamic_info.inputs + dynamic_info.outputs,
            )
        }
    torch.onnx.export(
        torch_model,  # model being run
        tuple(input_tensors),  # model input (or a tuple for multiple inputs)
        output_file_path,
        # where to save the model (can be a file or file-like object)
        export_params=True,
        # store the trained parameter weights inside the model file
        opset_version=ONNX_OPSET_VERSION,
        # the ONNX version to export the model to
        do_constant_folding=True,
        # whether to execute constant folding for optimization
        input_names=input_names,
        # the model's input names
        output_names=output_names,
        # the model's output names
        dynamic_axes=dynamic_info,
    )



import logging
import os
import warnings
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Tuple, Dict, Optional, Callable, Union, Sequence

import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm import tqdm

from nebullvm.api.frontend.utils import (
    check_inputs,
    ifnone,
    inspect_dynamic_size,
    QUANTIZATION_METRIC_MAP,
)
from nebullvm.base import (
    DeepLearningFramework,
    ModelParams,
    ModelCompiler,
    InputInfo,
    QuantizationType,
)
from nebullvm.converters import ONNXConverter
from nebullvm.optimizers.pytorch import PytorchBackendOptimizer
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.torch import (
    get_outputs_sizes_torch,
    create_model_inputs_torch,
    run_torch_model,
)
from nebullvm.inference_learners.base import PytorchBaseInferenceLearner
from nebullvm.measure import compute_optimized_running_time
from nebullvm.optimizers import ApacheTVMOptimizer, BaseOptimizer
from nebullvm.optimizers.multi_compiler import MultiCompilerOptimizer

logging.basicConfig(
    format="%(asctime)s %(message)s", datefmt="%d/%m/%Y %I:%M:%S %p"
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def _extract_dynamic_axis(
    torch_model: torch.nn.Module,
    dataloader: DataLoader,
    input_sizes: List[Tuple[int, ...]],
    batch_size: int,
    max_data: int = 100,
) -> Optional[Dict]:
    dynamic_axis = {"inputs": [{}] * len(input_sizes), "outputs": []}
    output_sizes = []
    for i, (input_tensors, y) in enumerate(dataloader):
        if i >= max_data:
            break
        inspect_dynamic_size(
            input_tensors, input_sizes, batch_size, dynamic_axis["inputs"]
        )
        outputs = tuple(run_torch_model(torch_model, input_tensors))
        if i == 0:
            dynamic_axis["outputs"] = [{}] * len(outputs)
            output_sizes = [tuple(output.shape[1:]) for output in outputs]
        inspect_dynamic_size(
            outputs, output_sizes, batch_size, dynamic_axis["outputs"]
        )
    if any(
        len(x) > 0 for x in (dynamic_axis["inputs"] + dynamic_axis["outputs"])
    ):
        return dynamic_axis
    return None


def _extract_info_from_data(
    model: torch.nn.Module,
    dataloader: Union[DataLoader, Sequence],
    batch_size: int,
    input_sizes: List[Tuple[int, ...]],
    input_types: List[str],
    dynamic_axis: Dict,
):
    input_row, _ = (
        dataloader[0]
        if isinstance(dataloader, Sequence)
        else next(iter(dataloader))
    )
    batch_size = ifnone(batch_size, int(input_row[0].shape[0]))
    input_sizes = ifnone(input_sizes, [tuple(x.shape[1:]) for x in input_row])
    input_types = ifnone(
        input_types,
        [
            "int" if isinstance(x.cpu(), torch.LongTensor) else "float"
            for x in input_row
        ],
    )
    dynamic_axis = ifnone(
        dynamic_axis,
        _extract_dynamic_axis(model, dataloader, input_sizes, batch_size),
    )
    return batch_size, input_sizes, input_types, dynamic_axis


def optimize_torch_model(
    model: torch.nn.Module,
    save_dir: str,
    dataloader: Union[DataLoader, Sequence] = None,
    batch_size: int = None,
    input_sizes: List[Tuple[int, ...]] = None,
    input_types: List[str] = None,
    extra_input_info: List[Dict] = None,
    use_torch_api: bool = False,
    dynamic_axis: Dict = None,
    perf_loss_ths: float = None,
    perf_metric: Union[str, Callable] = None,
    ignore_compilers: List[str] = None,
    custom_optimizers: List[BaseOptimizer] = None,
) -> PytorchBaseInferenceLearner:
    """Basic function for optimizing a torch model.

    This function saves the output model as well in a nebuly-readable format
    in order to avoid temporary-files corruptions which would prevent the model
    saving later in the process.

    Args:
        model (torch.nn.Module): Pytorch model that needs optimization.
        save_dir (str): Path to the directory where saving the final model.
        dataloader (DataLoader, optional): Data loader to be used for loading
            the user uploaded data. The data will be used for extracting all
            the data related information or completing the missing information
            in the case of partially given ones. If not given, both the
            `batch_size` and the `input_sizes` must be given by the user.
        batch_size (int, optional): The model batch size. Note that nebullvm
            does not support at the moment dynamic batch size, so a valid
            input should be given.
        input_sizes (List[Tuple]], optional): List containing the size of all
            the input tensors of the model. Note that even just a single
            tensor is needed as model input, this field must be a list
            containing (in the exposed case) a single element).
            The tuple must contain all the input tensor dimensions excluding
            the batch size. This means that the final input tensor size will
            be considered as (batch_size, *input_tensor_size)`, where
            `input_tensor_size` is one list element of `input_sizes`.
        input_types (List[str], optional): List of input types. If no value is
            given all the inputs will be considered as float type. The
            supported string values are "int" and "float".
        extra_input_info (List[Dict], optional): List of extra information
            needed for defining the input tensors, e.g. max_value and min_value
            the tensors can get.
        use_torch_api (bool): Parameter for using the torch api of compilers
            when available. The actual implementation supports only the torch
            interface for TVM. Note that when running the torch interface
            nebullvm will ignore the ONNX one once the torch implementation
            succeeds. Clearly, in case of failure of the torch API, a second
            tentative will be done with the ONNX interface.
        dynamic_axis (Dict, optional): Dictionary containing info about the
            dynamic axis. It should contain as keys both "inputs" and "outputs"
            and as values two lists of dictionaries where each dictionary
            represents the dynamic axis information for an input/output tensor.
            The inner dictionary should have as key an integer, i.e. the
            dynamic axis (considering also the batch size) and as value a
            string giving a "tag" to it, e.g. "batch_size".
        perf_loss_ths (float, optional): Tolerated relative error for
            performing approximation techniques before compiling the model.
            If no value is given, no optimization will be performed. Note that
            it will not be used for compilers using the torch API when
            `use_torch_api` is `True`. Just dynamic quantization will be
            performed, since no data is given as input.
        perf_metric (Union[Callable, str], optional): The metric to
            be used for accepting or refusing a precision-reduction
            optimization proposal. If none is given but a `perf_loss_ths` is
            received, the `nebullvm.measure.compute_relative_difference`
            metric will be used as default one. A user-defined metric can
            be passed as function accepting as inputs two tuples of tensors
            (produced by the baseline and the quantized model) and the related
            original labels.
            For more information see
            `nebullvm.measure.compute_relative_difference` and
            `nebullvm.measure.compute_accuracy_drop`. `perf_metric`
            accepts as value also a string containing the metric name. At the
            current stage the supported metrics are `"precision"` and
            `"accuracy"`.
        ignore_compilers (List[str], optional): List of DL compilers we want
            to ignore while running the optimization. Compiler name should be
            one between "tvm", "tensor RT", "openvino" and "onnxruntime".
        custom_optimizers (List[BaseOptimizer], optional): List of optimizers
            which can be used for producing InferenceLearners, i.e. models
            optimized for inference. This list is useful when some compilers,
            not included in the base version of nebullvm or not used
            by default, which are specific for a particular tasks, need
            to be used.

    Returns:
        PytorchBaseInferenceLearner: Optimized model usable with the classical
            Pytorch interface. Note that as a torch model it takes as input
            and it gives as output `torch.Tensor` s.
    """
    check_inputs(
        input_data=dataloader, batch_size=batch_size, input_sizes=input_sizes
    )
    if isinstance(perf_metric, str):
        perf_metric = QUANTIZATION_METRIC_MAP.get(perf_metric)
    if dataloader is not None:
        (
            batch_size,
            input_sizes,
            input_types,
            dynamic_axis,
        ) = _extract_info_from_data(
            model,
            dataloader,
            batch_size,
            input_sizes,
            input_types,
            dynamic_axis,
        )
        input_data = (
            DataManager.from_iterable(dataloader)
            if isinstance(dataloader, DataLoader)
            else DataManager(dataloader)
        )
    else:
        input_data = None
    if input_types is None:
        input_types = ["float"] * len(input_sizes)
    if extra_input_info is None:
        extra_input_info = [{}] * len(input_sizes)
    if not len(input_sizes) == len(input_types) == len(extra_input_info):
        raise ValueError(
            f"Mismatch in the input list lengths. Given {len(input_sizes)} "
            f"sizes, {len(input_types)} input types and "
            f"{len(extra_input_info)} extra input infos."
        )
    input_infos = [
        InputInfo(size=input_size, dtype=input_type, **extra_info)
        for input_size, input_type, extra_info in zip(
            input_sizes, input_types, extra_input_info
        )
    ]
    dl_library = DeepLearningFramework.PYTORCH
    model_params = ModelParams(
        batch_size=batch_size,
        input_infos=input_infos,
        output_sizes=get_outputs_sizes_torch(
            model,
            input_tensors=create_model_inputs_torch(batch_size, input_infos),
        ),
        dynamic_info=dynamic_axis,
    )
    ignore_compilers = (
        []
        if ignore_compilers is None
        else [ModelCompiler(compiler) for compiler in ignore_compilers]
    )
    input_tfms = MultiStageTransformation([])
    with TemporaryDirectory() as tmp_dir:
        logger.info("Running Optimization using torch interface (1/3)")
        if perf_loss_ths is not None:
            q_types = [
                None,
                QuantizationType.DYNAMIC,
                QuantizationType.HALF,
            ]
            if dataloader is not None:
                q_types.append(QuantizationType.STATIC)
        else:
            q_types = [None]
        torch_res = [
            _torch_api_optimization(
                model,
                model_params,
                perf_loss_ths,
                q_type,
                use_torch_api,
                input_data,
            )
            for q_type in tqdm(q_types)
        ]
        (torch_api_model, torch_api_latency, used_compilers,) = sorted(
            torch_res, key=lambda x: x[1]
        )[0]
        ignore_compilers.extend(used_compilers)
        logger.info("Running Optimization using ONNX interface (2/3)")
        model_converter = ONNXConverter()
        model_optimizer = MultiCompilerOptimizer(
            ignore_compilers=ignore_compilers,
            extra_optimizers=custom_optimizers,
            debug_mode=int(os.environ.get("DEBUG_MODE", "0")) > 0,
            logger=logger,
        )
        if model_optimizer.usable:
            onnx_path = model_converter.convert(
                model, model_params, Path(tmp_dir)
            )
            model_optimized = model_optimizer.optimize(
                model=str(onnx_path),
                output_library=dl_library,
                model_params=model_params,
                input_tfms=input_tfms,
                perf_loss_ths=perf_loss_ths,
                perf_metric=perf_metric,
                input_data=input_data,
            )
        else:
            model_optimized = None
        logger.info("Running comparison between optimized models (3/3).")
        model_optimized = _compare_optimized_models(
            model_optimized,
            torch_api_model,
            torch_api_latency,
        )
        if model_optimized is None:
            raise RuntimeError(
                "No valid compiled model has been produced. "
                "Look at the logs for further information about the failure."
            )
        model_optimized.save(save_dir)
    return model_optimized.load(save_dir)


def _get_optimizers_supporting_torch_api(
    use_extra_compilers: bool,
) -> List[Tuple[ModelCompiler, BaseOptimizer]]:
    optimizers = [
        (ModelCompiler.TORCHVISION, PytorchBackendOptimizer(logger=logger)),
    ]
    if use_extra_compilers:
        optimizers.append(
            (ModelCompiler.APACHE_TVM, ApacheTVMOptimizer(logger=logger))
        )
    return optimizers


def _torch_api_optimization(
    model: torch.nn.Module,
    model_params: ModelParams,
    quantization_ths: float,
    quantization_type: QuantizationType,
    use_extra_compilers: bool,
    input_data: DataManager,
) -> Tuple[Optional[PytorchBaseInferenceLearner], float, List]:
    used_compilers = []
    best_torch_opt_model = None
    best_latency = np.inf
    for compiler, optimizer in tqdm(
        _get_optimizers_supporting_torch_api(use_extra_compilers)
    ):
        try:
            if hasattr(optimizer, "optimize_from_torch"):
                candidate_model = optimizer.optimize_from_torch(
                    torch_model=model,
                    model_params=model_params,
                    perf_loss_ths=quantization_ths
                    if quantization_type is not None
                    else None,
                    quantization_type=quantization_type,
                    input_data=input_data,
                )
            else:
                candidate_model = optimizer.optimize(
                    model=model,
                    output_library=DeepLearningFramework.PYTORCH,
                    model_params=model_params,
                    perf_loss_ths=quantization_ths
                    if quantization_type is not None
                    else None,
                    quantization_type=quantization_type,
                    input_data=input_data,
                )
            candidate_latency = compute_optimized_running_time(candidate_model)
            if candidate_latency < best_latency:
                best_latency = candidate_latency
                best_torch_opt_model = candidate_model
            used_compilers.append(compiler)
        except Exception as ex:
            warnings.warn(
                f"Compilation failed with torch interface of {compiler}. "
                f"Got error {ex}. If possible the compilation will be "
                f"re-scheduled with the ONNX interface. Please consult the "
                f"documentation for further info or open an issue on GitHub "
                f"for receiving assistance."
            )
    return best_torch_opt_model, best_latency, used_compilers


def _compare_optimized_models(
    new_model: PytorchBaseInferenceLearner,
    previous_best_model: PytorchBaseInferenceLearner,
    previous_latency: float,
) -> PytorchBaseInferenceLearner:
    if new_model is not None:
        new_latency = compute_optimized_running_time(new_model)
        if new_latency < previous_latency:
            return new_model
    return previous_best_model



from typing import List, Tuple

import numpy as np
import torch

from nebullvm.config import NO_COMPILER_INSTALLATION

if torch.cuda.is_available():
    try:
        import polygraphy
        from tensorrt import IInt8EntropyCalibrator2
    except ImportError:
        import warnings

        if not NO_COMPILER_INSTALLATION:
            from nebullvm.installers.installers import install_tensor_rt

            warnings.warn(
                "No TensorRT valid installation has been found. "
                "Trying to install it from source."
            )
            install_tensor_rt()
            import polygraphy
            from tensorrt import IInt8EntropyCalibrator2
        else:
            warnings.warn(
                "No TensorRT valid installation has been found. "
                "It won't be possible to use it in the following steps."
            )
            IInt8EntropyCalibrator2 = object
else:
    IInt8EntropyCalibrator2 = object


class TensorRTCalibrator(IInt8EntropyCalibrator2):
    def __init__(
        self, batch_size: int, input_data: List[Tuple[np.ndarray, ...]]
    ):
        super(TensorRTCalibrator, self).__init__()
        self._bs = batch_size
        self._input_data = (x for x in input_data)

    def read_calibration_cache(self):
        return None

    def write_calibration_cache(self, cache):
        return None

    def get_batch(self, names):
        cuda_stream = polygraphy.Stream()
        try:
            data = next(self._input_data)
            cuda_data = [
                polygraphy.DeviceArray.copy_from(
                    input_tensor, stream=cuda_stream
                )
                for input_tensor in data
            ]
            return [input_tensor.ptr for input_tensor in cuda_data]
        except StopIteration:
            return None



from pathlib import Path
import subprocess
from typing import Optional, Callable

from nebullvm.base import DeepLearningFramework, ModelParams, QuantizationType
from nebullvm.inference_learners.openvino import (
    OPENVINO_INFERENCE_LEARNERS,
    OpenVinoInferenceLearner,
)
from nebullvm.optimizers.base import BaseOptimizer
from nebullvm.optimizers.quantization.openvino import quantize_openvino
from nebullvm.optimizers.quantization.utils import check_precision
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager
from nebullvm.utils.onnx import (
    get_input_names,
    create_model_inputs_onnx,
    run_onnx_model,
    convert_to_numpy,
)


class OpenVinoOptimizer(BaseOptimizer):
    """Class for compiling the AI models on Intel CPUs using OpenVino."""

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[OpenVinoInferenceLearner]:
        """Optimize the onnx model with OpenVino.

        Args:
            model (str): Path to the saved onnx model.
            output_library (str): DL Framework the optimized model will be
                compatible with.
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            OpenVinoInferenceLearner: Model optimized with OpenVino. The model
                will have an interface in the DL library specified in
                `output_library`.
        """
        self._log(
            f"Optimizing with {self.__class__.__name__} and "
            f"q_type: {quantization_type}."
        )
        cmd = [
            "mo",
            "--input_model",
            model,
            "--output_dir",
            str(Path(model).parent),
            "--input",
            ",".join(get_input_names(model)),
            "--input_shape",
            ",".join(
                [
                    f"{(model_params.batch_size,)+shape}"
                    for shape in model_params.input_sizes
                ]
            ),
        ]
        if (
            perf_loss_ths is not None
            and quantization_type is QuantizationType.HALF
        ):
            cmd = cmd + ["--data_type", "FP16"]
        elif (
            perf_loss_ths is not None
            and quantization_type is QuantizationType.DYNAMIC
        ):
            return None
        process = subprocess.Popen(cmd)
        process.wait()
        base_path = Path(model).parent
        openvino_model_path = base_path / f"{Path(model).stem}.xml"
        openvino_model_weights = base_path / f"{Path(model).stem}.bin"
        if (
            perf_loss_ths is not None
            and quantization_type is not QuantizationType.HALF
        ):
            if input_data is not None and quantization_type:
                input_data_onnx = input_data.get_numpy_list(300, with_ys=True)
            else:
                input_data_onnx = [
                    (
                        create_model_inputs_onnx(
                            model_params.batch_size, model_params.input_infos
                        ),
                        0,
                    )
                ]
            # Add post training optimization
            openvino_model_path, openvino_model_weights = quantize_openvino(
                model_topology=str(openvino_model_path),
                model_weights=str(openvino_model_weights),
                input_names=get_input_names(model),
                input_data=input_data_onnx,
            )
        learner = OPENVINO_INFERENCE_LEARNERS[output_library].from_model_name(
            model_name=str(openvino_model_path),
            model_weights=str(openvino_model_weights),
            network_parameters=model_params,
            input_tfms=input_tfms,
        )
        if perf_loss_ths is not None:
            if input_data is None:
                inputs = [learner.get_inputs_example()]
                ys = None
            else:
                inputs, ys = input_data.get_list(
                    100, with_ys=True, shuffle=True
                )
            output_data_onnx = [
                tuple(
                    run_onnx_model(
                        model,
                        [convert_to_numpy(x) for x in tuple_],
                    )
                )
                for tuple_ in inputs
            ]
            is_valid = check_precision(
                learner,
                inputs,
                output_data_onnx,
                perf_loss_ths,
                metric_func=perf_metric,
                ys=ys,
            )
            if not is_valid:
                return None
        return learner



from nebullvm.optimizers.base import BaseOptimizer  # noqa F401
from nebullvm.optimizers.onnx import ONNXOptimizer  # noqa F401
from nebullvm.optimizers.openvino import OpenVinoOptimizer  # noqa F401
from nebullvm.optimizers.tensor_rt import TensorRTOptimizer  # noqa F401
from nebullvm.optimizers.tvm import ApacheTVMOptimizer  # noqa F401

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import logging
from abc import abstractmethod, ABC
from logging import Logger
from typing import Optional, Callable, Any

from nebullvm.base import DeepLearningFramework, ModelParams, QuantizationType
from nebullvm.inference_learners.base import BaseInferenceLearner
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager


class BaseOptimizer(ABC):
    """Base class for Optimizers"""

    def __init__(self, logger: Logger = None):
        self.logger = logger

    @abstractmethod
    def optimize(
        self,
        model: Any,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> Optional[BaseInferenceLearner]:
        raise NotImplementedError

    def _log(self, message: str, level: int = logging.INFO):
        if self.logger is None:
            logging.log(level, message)
        else:
            self.logger.log(level, message)



import json
import warnings
from logging import Logger
from pathlib import Path
from typing import Dict, Type, Tuple, Callable, List
import uuid

import cpuinfo
import numpy as np
import torch
from tqdm import tqdm

from nebullvm.base import (
    ModelCompiler,
    DeepLearningFramework,
    ModelParams,
    QuantizationType,
)
from nebullvm.config import NEBULLVM_DEBUG_FILE
from nebullvm.inference_learners.base import BaseInferenceLearner
from nebullvm.measure import compute_optimized_running_time
from nebullvm.optimizers import (
    BaseOptimizer,
    TensorRTOptimizer,
    ApacheTVMOptimizer,
    OpenVinoOptimizer,
    ONNXOptimizer,
)
from nebullvm.transformations.base import MultiStageTransformation
from nebullvm.utils.data import DataManager

COMPILER_TO_OPTIMIZER_MAP: Dict[ModelCompiler, Type[BaseOptimizer]] = {
    ModelCompiler.APACHE_TVM: ApacheTVMOptimizer,
    ModelCompiler.OPENVINO: OpenVinoOptimizer,
    ModelCompiler.TENSOR_RT: TensorRTOptimizer,
    ModelCompiler.ONNX_RUNTIME: ONNXOptimizer,
}


def _tvm_is_available() -> bool:
    try:
        import tvm  # noqa F401

        return True
    except ImportError:
        return False


def select_compilers_from_hardware():
    compilers = [ModelCompiler.ONNX_RUNTIME]
    if _tvm_is_available():
        compilers.append(ModelCompiler.APACHE_TVM)
    if torch.cuda.is_available():
        compilers.append(ModelCompiler.TENSOR_RT)
    cpu_raw_info = cpuinfo.get_cpu_info()["brand_raw"].lower()
    if "intel" in cpu_raw_info:
        compilers.append(ModelCompiler.OPENVINO)
    return compilers


def _optimize_with_compiler(
    compiler: ModelCompiler,
    logger: Logger,
    metric_func: Callable = None,
    **kwargs,
) -> Tuple[BaseInferenceLearner, float]:
    optimizer = COMPILER_TO_OPTIMIZER_MAP[compiler](logger)
    return _optimize_with_optimizer(optimizer, logger, metric_func, **kwargs)


def _save_info(
    optimizer: BaseOptimizer,
    score: float,
    debug_file: str,
    optimization_params: Dict,
):
    if Path(debug_file).exists():
        with open(debug_file, "r") as f:
            old_dict = json.load(f)
    else:
        old_dict = {}
    optimization_string = optimizer.__class__.__name__
    quantization_string = "_".join(
        [
            str(optimization_params.get(param)) or ""
            for param in ["perf_loss_ths", "quantization_type"]
        ]
    )
    if len(quantization_string) > 1:
        optimization_string += "_" + quantization_string
    old_dict[optimization_string] = f"{score}"
    with open(debug_file, "w") as f:
        json.dump(old_dict, f)


def _optimize_with_optimizer(
    optimizer: BaseOptimizer,
    logger: Logger,
    metric_func: Callable = None,
    debug_file: str = None,
    **kwargs,
) -> Tuple[BaseInferenceLearner, float]:
    if metric_func is None:
        metric_func = compute_optimized_running_time
    try:
        model_optimized = optimizer.optimize(**kwargs)
        latency = metric_func(model_optimized)
    except Exception as ex:
        warning_msg = (
            f"Compilation failed with {optimizer.__class__.__name__}. "
            f"Got error {ex}. The optimizer will be skipped."
        )
        if logger is None:
            warnings.warn(warning_msg)
        else:
            logger.warning(warning_msg)
        latency = np.inf
        model_optimized = None
    if debug_file:
        _save_info(optimizer, latency, debug_file, kwargs)
    return model_optimized, latency


class MultiCompilerOptimizer(BaseOptimizer):
    """Run all the optimizers available for the given hardware and select the
    best optimized model in terms of either latency or user defined
    performance.

    Attributes:
        logger (Logger, optional): User defined logger.
        ignore_compilers (List[str], optional): List of compilers that must
            be ignored.
        extra_optimizers (List[BaseOptimizer], optional): List of optimizers
            defined by the user. It usually contains optimizers specific for
            user-defined tasks or optimizers built for the specific model.
            Note that, if given, the optimizers must be already initialized,
            i.e. they could have a different Logger than the one defined in
            `MultiCompilerOptimizer`.
        debug_mode (bool, optional): Boolean flag for activating the debug
            mode. When activated, all the performances of the the different
            containers  will be stored in a json file saved in the working
            directory. Default is False.
    """

    def __init__(
        self,
        logger: Logger = None,
        ignore_compilers: List = None,
        extra_optimizers: List[BaseOptimizer] = None,
        debug_mode: bool = False,
    ):
        super().__init__(logger)
        self.compilers = [
            compiler
            for compiler in select_compilers_from_hardware()
            if compiler not in (ignore_compilers or [])
        ]
        self.extra_optimizers = extra_optimizers
        self.debug_file = (
            f"{uuid.uuid4()}_{NEBULLVM_DEBUG_FILE}" if debug_mode else None
        )

    def optimize(
        self,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
    ) -> BaseInferenceLearner:
        """Optimize the ONNX model using the available compilers.

        Args:
            model (str): Path to the ONNX model.
            output_library (DeepLearningFramework): Framework of the optimized
                model (either torch on tensorflow).
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            BaseInferenceLearner: Model optimized for inference.
        """
        if perf_loss_ths is not None and quantization_type is None:
            quantization_types = [
                None,
                QuantizationType.DYNAMIC,
                QuantizationType.HALF,
            ]
            if input_data is not None:
                quantization_types.append(QuantizationType.STATIC)
        else:
            quantization_types = [quantization_type]
        optimized_models = [
            _optimize_with_compiler(
                compiler,
                logger=self.logger,
                model=model,
                output_library=output_library,
                model_params=model_params,
                input_tfms=input_tfms.copy()
                if input_tfms is not None
                else None,
                debug_file=self.debug_file,
                perf_loss_ths=perf_loss_ths if q_type is not None else None,
                quantization_type=q_type,
                perf_metric=perf_metric,
                input_data=input_data,
            )
            for compiler in self.compilers
            for q_type in tqdm(quantization_types)
        ]
        if self.extra_optimizers is not None:
            self._log("Running extra-optimizers...")
            optimized_models += [
                _optimize_with_optimizer(
                    op,
                    logger=self.logger,
                    model=model,
                    output_library=output_library,
                    model_params=model_params,
                    input_tfms=input_tfms.copy()
                    if input_tfms is not None
                    else None,
                    debug_file=self.debug_file,
                    perf_loss_ths=perf_loss_ths
                    if q_type is not None
                    else None,
                    quantization_type=q_type,
                    perf_metric=perf_metric,
                    input_data=input_data,
                )
                for op in self.extra_optimizers
                for q_type in tqdm(quantization_types)
            ]
        optimized_models.sort(key=lambda x: x[1], reverse=False)
        return optimized_models[0][0]

    def optimize_on_custom_metric(
        self,
        metric_func: Callable,
        model: str,
        output_library: DeepLearningFramework,
        model_params: ModelParams,
        input_tfms: MultiStageTransformation = None,
        perf_loss_ths: float = None,
        quantization_type: QuantizationType = None,
        perf_metric: Callable = None,
        input_data: DataManager = None,
        return_all: bool = False,
    ):
        """Optimize the ONNX model using the available compilers and give the
        best result sorting by user-defined metric.

        Args:
            metric_func (Callable): function which should be used for sorting
                the compiled models. The metric_func should take as input an
                InferenceLearner and return a numerical value. Note that the
                outputs will be sorted in an ascendant order, i.e. the compiled
                model with the smallest value will be selected.
            model (str): Path to the ONNX model.
            output_library (DeepLearningFramework): Framework of the optimized
                model (either torch on tensorflow).
            model_params (ModelParams): Model parameters.
            input_tfms (MultiStageTransformation, optional): Transformations
                to be performed to the model's input tensors in order to
                get the prediction.
            return_all (bool, optional): Boolean flag. If true the method
                returns the tuple (compiled_model, score) for each available
                compiler. Default `False`.
            perf_loss_ths (float, optional): Threshold for the accepted drop
                in terms of precision. Any optimized model with an higher drop
                will be ignored.
            quantization_type (QuantizationType, optional): The desired
                quantization algorithm to be used.
            perf_metric (Callable, optional): If given it should
                compute the difference between the quantized and the normal
                prediction.
            input_data (DataManager, optional): User defined data.

        Returns:
            Union[BaseInferenceLearner, Tuple[BaseInferenceLearner, float]]:
                The method returns just a model optimized for inference if
                `return_all` is `False` or all the compiled models and their
                scores otherwise.
        """
        if perf_loss_ths is not None and quantization_type is None:
            quantization_types = [
                None,
                QuantizationType.DYNAMIC,
                QuantizationType.HALF,
            ]
            if input_data is not None:
                quantization_types.append(QuantizationType.STATIC)
        else:
            quantization_types = [quantization_type]
        optimized_models = [
            _optimize_with_compiler(
                compiler,
                metric_func=metric_func,
                logger=self.logger,
                model=model,
                output_library=output_library,
                model_params=model_params,
                input_tfms=input_tfms.copy()
                if input_tfms is not None
                else None,
                debug_file=self.debug_file,
                perf_loss_ths=perf_loss_ths if q_type is not None else None,
                quantization_type=q_type,
                perf_metric=perf_metric,
                input_data=input_data,
            )
            for compiler in self.compilers
            for q_type in tqdm(quantization_types)
        ]
        if self.extra_optimizers is not None:
            optimized_models += [
                _optimize_with_optimizer(
                    op,
                    logger=self.logger,
                    model=model,
                    output_library=output_library,
                    model_params=model_params,
                    input_tfms=input_tfms.copy()
                    if input_tfms is not None
                    else None,
                    debug_file=self.debug_file,
                    perf_loss_ths=perf_loss_ths
                    if q_type is not None
                    else None,
                    quantization_type=q_type,
                    perf_metric=perf_metric,
                    input_data=input_data,
                )
                for op in self.extra_optimizers
                for q_type in tqdm(quantization_types)
            ]
        if return_all:
            return optimized_models
        optimized_models.sort(key=lambda x: x[1], reverse=False)
        return optimized_models[0][0]

    @property
    def usable(self) -> bool:
        return len(self.compilers) > 0 or (
            self.extra_optimizers is not None
            and len(self.extra_optimizers) > 0
        )



from pathlib import Path
from setuptools import setup, find_packages

REQUIREMENTS = [
    "numpy>=1.19.0",
    "onnx>=1.10.0",
    "onnxmltools>=1.11.0",
    "py-cpuinfo>=8.0.0",
    "tensorflow>=2.7.0, <2.8.0",
    "tf2onnx>=1.8.4",
    "torch>=1.10.0",
    "tqdm>=4.63.0",
]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="nebullvm",
    version="0.3.1",
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    package_data={
        # Include all bash files:
        "": ["*.sh"],
        # And include any file needed for config
        "nebullvm": ["*config.cmake"],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
)
